KEY_BINDINGS = {
    'PREVIOUS_MESSAGE': {
        'keys': {'k', 'up'},
        'help_text': 'Previous message',

    },
    'NEXT_MESSAGE': {
        'keys': {'j', 'down'},
        'help_text': 'Next message',
    },
    'GO_LEFT': {
        'keys': {'h', 'left'},
        'help_text': 'Go left',
    },
    'GO_RIGHT': {
        'keys': {'l', 'right'},
        'help_text': 'Go right',
    },
    'SCROLL_TO_TOP': {
        'keys': {'K', 'page up'},
        'help_text': 'Scroll to top',
    },
    'SCROLL_TO_BOTTOM': {
        'keys': {'J', 'page down'},
        'help_text': 'Scroll to bottom',
    },
    'END_MESSAGE': {
        'keys': {'G', 'end'},
        'help_text': 'Go to last message in view',
    },
    'REPLY_MESSAGE': {
        'keys': {'r'},
        'help_text': 'Reply to a message',
    },
    'REPLY_AUTHOR': {
        'keys': {'R'},
        'help_text': 'Reply to a author',
    },
    'STREAM_MESSAGE': {
        'keys': {'c'},
        'help_text': 'New stream message',
    },
    'GO_BACK': {
        'keys': {'esc'},
        'help_text': 'Go Back',
    },
    'STREAM_NARROW': {
        'keys': {'S'},
        'help_text': 'Narrow to a stream',
    },
    'TOPIC_NARROW': {
        'keys': {'s'},
        'help_text': 'Narrow to a topic',
    },
    'NEXT_UNREAD_TOPIC': {
        'keys': {'n'},
        'help_text': 'Next unread topic',
    },
    'NEXT_UNREAD_PM': {
        'keys': {'p'},
        'help_text': 'Next unread private message',
    },
    'SEND_MESSAGE': {
        'keys': {'meta enter'},
        'help_text': 'Send a message',
    },
    'SEARCH_PEOPLE': {
        'keys': {'w'},
        'help_text': 'Search People',
    },
    'SEARCH_MESSAGES': {
        'keys': {'/'},
        'help_text': 'Search Messages',
    },
    'SEARCH_STREAMS': {
        'keys': {'q'},
        'help_text': 'Search Streams',
    },
    'HELP': {
        'keys': {'?'},
        'help_text': 'Display help menu',
    },
    'ENTER': {
        'keys': {'enter'},
        'help_text': 'Perform current action',
    },
    'ALL_PM': {
        'keys': {'P'},
        'help_text': 'Show all private messgaes',
    },
    'QUIT_HELP': {
        'keys': {'q', 'esc'},
        'help_text': 'Quit help menu',
    },
    'PRIVATE_MESSAGE': {
        'keys': {'x'},
        'help_text': 'New private message',
    },
    'TAB': {
        'keys': {'tab'},
        'help_text': 'Toggle focus box in compose box'
    }
}


def is_command_key(command: str, key: str) -> bool:
    """
    Returns the mapped binding for a key if mapped
    or the key otherwise.
    """
    if key in KEY_BINDINGS[command]['keys']:
        return True
    else:
        return False



import platform
import re
from urllib.parse import urlparse
from typing import Any, Tuple

import urwid

from zulipterminal.config import is_command_key
from zulipterminal.ui_tools.boxes import WriteBox, SearchBox
from zulipterminal.ui_tools.views import (
    RightColumnView,
    MiddleColumnView,
    StreamsView,
    LeftColumnView,
)


class View(urwid.WidgetWrap):
    """
    A class responsible for providing the application's interface.
    """
    palette = {
        'default': [
            (None,           'white',         'black'),
            ('selected',     'light magenta', 'dark blue'),
            ('msg_selected', 'light green',   'black'),
            ('header',       'dark cyan',     'dark blue',  'bold'),
            ('custom',       'white',         'dark blue',  'underline'),
            ('content',      'white',         'black',      'standout'),
            ('name',         'yellow, bold',  'black'),
            ('unread',       'light blue',    'black'),
            ('active',       'white',         'black'),
            ('idle',         'yellow',        'black'),
            ('title',        'white, bold',   'black'),
            ('time',         'light blue', 'black'),
            ('bar',          'white',         'black'),
            ('emoji',        'light magenta', 'black'),
        ],
        'light': [
            (None,           'black',        'white'),
            ('selected',     'white',        'dark blue'),
            ('msg_selected', 'dark blue',    'light gray'),
            ('header',       'white',        'dark blue',  'bold'),
            ('custom',       'white',        'dark blue',  'underline'),
            ('content',      'black',        'light gray', 'standout'),
            ('name',         'dark magenta', 'light gray', 'bold'),
        ],
        'blue': [
            (None,           'black',        'light blue'),
            ('selected',     'white',        'dark blue'),
            ('msg_selected', 'black',        'light gray'),
            ('header',       'black',        'dark blue',  'bold'),
            ('custom',       'white',        'dark blue',  'underline'),
            ('content',      'black',        'light gray', 'standout'),
            ('name',         'dark red',     'light gray', 'bold'),
        ]
    }

    def __init__(self, controller: Any) -> None:
        self.controller = controller
        self.model = controller.model
        self.client = controller.client
        self.users = self.model.users
        self.streams = self.model.streams
        self.write_box = WriteBox(self)
        self.search_box = SearchBox(self.controller)
        super(View, self).__init__(self.main_window())

    def left_column_view(self) -> Any:
        self.left_col_w = LeftColumnView(self)
        return self.left_col_w

    def message_view(self) -> Any:
        self.middle_column = MiddleColumnView(self.model, self.write_box,
                                              self.search_box)
        w = urwid.LineBox(self.middle_column, bline="")
        return w

    def right_column_view(self) -> Any:
        self.users_view = RightColumnView(self)
        w = urwid.LineBox(
            self.users_view, title=u"Users",
            tlcorner=u'─', tline=u'─', lline=u'',
            trcorner=u'─', blcorner=u'─', rline=u'',
            bline=u'', brcorner=u''
        )
        return w

    def main_window(self) -> Any:
        left_column = self.left_column_view()
        center_column = self.message_view()
        right_column = self.right_column_view()
        body = [
            (25, left_column),
            ('weight', 10, center_column),
            (25, right_column),
        ]
        self.body = urwid.Columns(body, focus_column=1)

        div_char = '═'
        profile = self.controller.client.get_profile()

        base_url = '{uri.scheme}://{uri.netloc}/'.format(
                uri=urlparse(self.controller.client.base_url))

        title_text = " {full_name} ({email}) - {server} ".format(
                server=base_url, **profile)
        title_bar = urwid.Columns([
            urwid.Divider(div_char=div_char),
            (len(title_text), urwid.Text([title_text])),
            urwid.Divider(div_char=div_char),
        ])

        w = urwid.Frame(self.body, title_bar, focus_part='body')
        return w

    def keypress(self, size: Tuple[int, int], key: str) -> str:
        if self.controller.editor_mode:
            return self.controller.editor.keypress((size[1],), key)
        # Redirect commands to message_view.
        elif is_command_key('SEARCH_MESSAGES', key) or\
                is_command_key('NEXT_UNREAD_TOPIC', key) or\
                is_command_key('NEXT_UNREAD_PM', key) or\
                is_command_key('PRIVATE_MESSAGE', key):
            self.body.focus_col = 1
            self.middle_column.keypress(size, key)
            return key
        elif is_command_key('SEARCH_PEOPLE', key):
            # Start User Search if not in editor_mode
            self.users_view.keypress(size, 'w')
            self.body.focus_col = 2
            self.user_search.set_edit_text("")
            self.controller.editor_mode = True
            self.controller.editor = self.user_search
            return key
        elif is_command_key('SEARCH_STREAMS', key):
            # jump stream search
            self.left_col_w.keypress(size, 'q')
            self.body.focus_col = 0
            self.stream_w.search_box.set_edit_text("")
            self.controller.editor_mode = True
            self.controller.editor = self.stream_w.search_box
            return key
        elif is_command_key('HELP', key):
            # Show help menu
            self.controller.show_help()
            return key
        # replace alternate keys with arrow/functional keys
        # This is needed for navigating in widgets
        # other than message_view.
        elif is_command_key('PREVIOUS_MESSAGE', key):
            key = 'up'
        elif is_command_key('NEXT_MESSAGE', key):
            key = 'down'
        elif is_command_key('GO_LEFT', key):
            key = 'left'
        elif is_command_key('GO_RIGHT', key):
            key = 'right'
        elif is_command_key('SCROLL_TO_TOP', key):
            key = 'page up'
        elif is_command_key('SCROLL_TO_BOTTOM', key):
            key = 'page down'
        elif is_command_key('END_MESSAGE', key):
            key = 'end'
        return super(View, self).keypress(size, key)


class Screen(urwid.raw_display.Screen):

    def write(self, data: Any) -> None:
        if "Microsoft" in platform.platform():
            # replace urwid's SI/SO, which produce artifacts under WSL.
            # https://github.com/urwid/urwid/issues/264#issuecomment-358633735
            # Above link describes the change.
            data = re.sub("[\x0e\x0f]", "", data)
        super().write(data)



import json
from threading import Thread
import time
from typing import Any, Dict, List, FrozenSet, Set, Union, Optional

import urwid

from zulipterminal.helper import (
    asyncn,
    classify_unread_counts,
    index_messages,
    set_count
)
from zulipterminal.ui_tools.utils import create_msg_box_list


class Model:
    """
    A class responsible for storing the data to be displayed.
    """

    def __init__(self, controller: Any) -> None:
        self.controller = controller
        self.client = controller.client
        # Get message after registering to the queue.
        self.msg_view = None  # type: Any
        self.anchor = 0
        self.num_before = 30
        self.num_after = 10
        self.msg_list = None  # type: Any
        self.narrow = []  # type: List[Any]
        self.update = False
        self.stream_id = -1
        self.stream_dict = {}  # type: Dict[int, Any]
        self.recipients = frozenset()  # type: FrozenSet[Any]
        self.index = None  # type: Any
        self.user_id = -1  # type: int
        self.initial_data = {}  # type: Dict[str, Any]
        self._update_user_id()
        self._update_initial_data()
        self.users = self.get_all_users()
        self.muted_streams = list()  # type: List[int]
        self.streams = self.get_subscribed_streams()
        self.muted_topics = self.initial_data['muted_topics']
        self.unread_counts = classify_unread_counts(self)

    @asyncn
    def _update_user_id(self) -> None:
        self.user_id = self.client.get_profile()['user_id']

    def _update_realm_users(self) -> None:
        self.initial_data['realm_users'] = self.client.get_members(
            request={
                'client_gravatar': True,
            }
        )['members']

    def get_focus_in_current_narrow(self) -> Union[int, Set[None]]:
        """
        Returns the focus in the current narrow.
        For no existing focus this returns {}, otherwise the message ID.
        """
        return self.index['pointer'][str(self.narrow)]

    def set_focus_in_current_narrow(self, focus_message: int) -> None:
        self.index['pointer'][str(self.narrow)] = focus_message

    def set_narrow(self, *,
                   stream: Optional[str]=None, topic: Optional[str]=None,
                   search: Optional[str]=None,
                   pm_with: Optional[str]=None) -> bool:
        if search and not(stream or topic or pm_with):
            new_narrow = [['search', search]]
        elif stream and topic and not(search or pm_with):
            new_narrow = [["stream", stream],
                          ["topic", topic]]
        elif stream and not(topic or search or pm_with):
            new_narrow = [['stream', stream]]
        elif pm_with == '' and not(stream or topic or search):
            new_narrow = [['is', 'private']]
        elif pm_with and not(stream or topic or search):
            new_narrow = [['pm_with', pm_with]]
        elif not stream and not topic and not search and not pm_with:
            new_narrow = []
        else:
            raise RuntimeError("Model.set_narrow parameters used incorrectly.")

        if new_narrow != self.narrow:
            self.narrow = new_narrow
            return False
        else:
            return True

    def get_message_ids_in_current_narrow(self) -> Set[int]:
        narrow = self.narrow
        if narrow == []:
            current_ids = self.index['all_messages']
        elif narrow[0][0] == 'stream':
            stream_id = self.stream_id
            if len(narrow) == 1:
                current_ids = self.index['all_stream'][stream_id]
            elif len(narrow) == 2:
                topic = narrow[1][1]
                current_ids = self.index['stream'][stream_id][topic]
        elif narrow[0][1] == 'private':
            current_ids = self.index['all_private']
        elif narrow[0][0] == 'pm_with':
            recipients = self.recipients
            current_ids = self.index['private'][recipients]
        elif narrow[0][0] == 'search':
            current_ids = self.index['search']
        return current_ids.copy()

    def get_messages(self, first_anchor: bool) -> Any:
        request = {
            'anchor': self.anchor,
            'num_before': self.num_before,
            'num_after': self.num_after,
            'apply_markdown': False,
            'use_first_unread_anchor': first_anchor,
            'client_gravatar': False,
            'narrow': json.dumps(self.narrow),
        }
        response = self.client.do_api_query(request, '/json/messages',
                                            method="GET")
        if response['result'] == 'success':
            self.index = index_messages(response['messages'], self, self.index)
            if first_anchor:
                self.index[str(self.narrow)] = response['anchor']
            query_range = self.num_after + self.num_before + 1
            if len(response['messages']) < (query_range):
                self.update = True
            return self.index

    def _update_initial_data(self) -> None:
        try:
            # Thread Processes to reduces start time.
            get_messages = Thread(target=self.get_messages,
                                  kwargs={'first_anchor': True})
            get_messages.start()
            update_realm_users = Thread(target=self._update_realm_users)
            update_realm_users.start()
            result = self.client.register(
                fetch_event_types=[
                    'presence',
                    'subscription',
                    'message',
                    'update_message_flags',
                    'muted_topics',
                ],
                client_gravatar=True,
            )
            self.initial_data.update(result)
            # Join process to ensure they are completed
            update_realm_users.join()
            get_messages.join()
        except Exception:
            print("Invalid API key")
            raise urwid.ExitMainLoop()

    def get_all_users(self) -> List[Dict[str, Any]]:
        # Dict which stores the active/idle status of users (by email)
        presences = self.initial_data['presences']

        # Construct a dict of each user in the realm to look up by email
        # and a user-id to email mapping
        self.user_dict = dict()  # type: Dict[str, Dict[str, Any]]
        self.user_id_email_dict = dict()  # type: Dict[int, str]
        for user in self.initial_data['realm_users']:
            email = user['email']
            if email in presences:  # presences currently subset of all users
                status = presences[email]['aggregated']['status']
            else:
                # TODO: Consider if bots & other no-presence results should
                # also really be treated as 'idle' and adjust accordingly
                status = 'idle'
            self.user_dict[email] = {
                'full_name': user['full_name'],
                'email': email,
                'user_id': user['user_id'],
                'status': status,
            }
            self.user_id_email_dict[user['user_id']] = email

        # Generate filtered lists for active & idle users
        active = [properties for properties in self.user_dict.values()
                  if properties['status'] == 'active']
        idle = [properties for properties in self.user_dict.values()
                if properties['status'] == 'idle']

        # Construct user_list from sorted components of each list
        user_list = sorted(active, key=lambda u: u['full_name'])
        user_list += sorted(idle, key=lambda u: u['full_name'])

        return user_list

    def get_subscribed_streams(self) -> List[List[str]]:
        subscriptions = self.initial_data['subscriptions']
        # Store streams in id->Stream format
        for stream in subscriptions:
            self.stream_dict[stream['stream_id']] = stream
            # Add if stream is muted.
            if stream['in_home_view'] is False:
                self.muted_streams.append(stream['stream_id'])

        stream_names = [[
            stream['name'],
            stream['stream_id'],
            stream['color'],
        ] for stream in subscriptions
        ]
        return sorted(stream_names, key=lambda s: s[0].lower())

    def append_message(self, response: Dict[str, Any]) -> None:
        """
        Adds message to the end of the view.
        """
        response['flags'] = []
        if hasattr(self.controller, 'view') and self.update:
            self.index = index_messages([response], self, self.index)
            msg_w_list = create_msg_box_list(self, [response['id']])
            if not msg_w_list:
                return
            else:
                msg_w = msg_w_list[0]
            if not self.narrow:
                self.msg_list.log.append(msg_w)

            elif self.narrow[0][1] == response['type'] and\
                    len(self.narrow) == 1:
                self.msg_list.log.append(msg_w)

            elif response['type'] == 'stream' and len(self.narrow) == 2 and\
                    self.narrow[1][1] == response['subject']:
                self.msg_list.log.append(msg_w)

            elif response['type'] == 'private' and len(self.narrow) == 1 and\
                    self.narrow[0][0] == "pm_with":
                recipients = self.recipients
                msg_recipients = frozenset([
                    self.user_id,
                    self.user_dict[self.narrow[0][1]]['user_id']
                ])
                if recipients == msg_recipients:
                    self.msg_list.log.append(msg_w)

            set_count([response['id']], self.controller, 1)
            self.controller.update_screen()

    def update_message(self, response: Dict[str, Any]) -> None:
        """
        Updates previously rendered message.
        """
        message_id = response['message_id']
        content = response['content']
        # If the message is indexed
        if self.index['messages'][message_id] != {}:
            message = self.index['messages'][message_id]
            message['content'] = content
            self.index['messages'][message_id] = message
            self.update_rendered_view(message_id)

    def update_reaction(self, response: Dict[str, Any]) -> None:
        message_id = response['message_id']
        # If the message is indexed
        if self.index['messages'][message_id] != {}:

            message = self.index['messages'][message_id]
            if response['op'] == 'add':
                message['reactions'].append(
                    {
                        'user': response['user'],
                        'reaction_type': response['reaction_type'],
                        'emoji_code': response['emoji_code'],
                        'emoji_name': response['emoji_name'],
                    }
                )
            else:
                emoji_code = response['emoji_code']
                for reaction in message['reactions']:
                    # Since Who reacted is not displayed,
                    # remove the first one encountered
                    if reaction['emoji_code'] == emoji_code:
                        message['reactions'].remove(reaction)

            self.index['messages'][message_id] = message
            self.update_rendered_view(message_id)

    def update_rendered_view(self, msg_id: int) -> None:
        # Update new content in the rendered view
        for msg_w in self.msg_list.log:
            if msg_w.original_widget.message['id'] == msg_id:
                msg_w_list = create_msg_box_list(self, [msg_id])
                if not msg_w_list:
                    return
                else:
                    new_msg_w = msg_w_list[0]
                msg_pos = self.msg_list.log.index(msg_w)
                self.msg_list.log[msg_pos] = new_msg_w
                self.controller.update_screen()

    @asyncn
    def poll_for_events(self) -> None:
        queue_id = self.controller.queue_id
        last_event_id = self.controller.last_event_id
        while True:
            if queue_id is None:
                self.controller.register_initial_desired_events()
                queue_id = self.controller.queue_id
                last_event_id = self.controller.last_event_id

            response = self.client.get_events(
                queue_id=queue_id,
                last_event_id=last_event_id
            )

            if 'error' in response['result']:
                if response["msg"].startswith("Bad event queue id:"):
                    # Our event queue went away, probably because
                    # we were asleep or the server restarted
                    # abnormally.  We may have missed some
                    # events while the network was down or
                    # something, but there's not really anything
                    # we can do about it other than resuming
                    # getting new ones.
                    #
                    # Reset queue_id to register a new event queue.
                    queue_id = None
                time.sleep(1)
                continue
            for event in response['events']:
                last_event_id = max(last_event_id, int(event['id']))
                if event['type'] == 'message':
                    self.append_message(event['message'])
                if event['type'] == 'update_message':
                    # FIXME: Support Topic Editing
                    if 'subject' in event.keys():
                        continue
                    else:
                        self.update_message(event)
                if event['type'] == 'reaction':
                    self.update_reaction(event)



from platform import platform
from typing import Any, List
import os
import sys
import time

import urwid
import zulip

from zulipterminal.helper import asyncn
from zulipterminal.model import Model
from zulipterminal.ui import View, Screen
from zulipterminal.ui_tools.utils import create_msg_box_list
from zulipterminal.ui_tools.views import HelpView


class Controller:
    """
    A class responsible for setting up the model and view and running
    the application.
    """

    def __init__(self, config_file: str, theme: str) -> None:
        self.show_loading()
        self.client = zulip.Client(config_file=config_file,
                                   client='ZulipTerminal/0.1.0 ' + platform())
        # Register to the queue before initializing Model or View
        # so that we don't lose any updates while messages are being fetched.
        self.register_initial_desired_events()
        self.model = Model(self)
        self.view = View(self)
        # Start polling for events after view is rendered.
        self.model.poll_for_events()
        self.theme = theme
        self.editor_mode = False  # type: bool
        self.editor = None  # type: Any

    @asyncn
    def show_loading(self) -> None:

        def spinning_cursor() -> Any:
            while True:
                for cursor in '|/-\\':
                    yield cursor

        spinner = spinning_cursor()
        sys.stdout.write("\033[92mWelcome to Zulip.\033[0m\n")
        while not hasattr(self, 'view'):
            next_spinner = "Loading " + next(spinner)
            sys.stdout.write(next_spinner)
            sys.stdout.flush()
            time.sleep(0.1)
            sys.stdout.write('\b'*len(next_spinner))

        sys.stdout.write('\n')
        self.capture_stdout()

    def capture_stdout(self, path: str='debug.log') -> None:
        if hasattr(self, '_stdout'):
            return

        self._stdout = sys.stdout
        sys.stdout = open(path, 'a')

    def restore_stdout(self) -> None:
        if not hasattr(self, '_stdout'):
            return

        sys.stdout.flush()
        sys.stdout.close()
        sys.stdout = self._stdout
        sys.stdout.write('\n')
        del self._stdout

    def update_screen(self) -> None:
        # Write something to update pipe to trigger draw_screen
        os.write(self.update_pipe, b'1')

    def draw_screen(self, *args: Any, **kwargs: Any) -> None:
        self.loop.draw_screen()

    def show_help(self) -> None:
        self.loop.widget = urwid.LineBox(urwid.Overlay(
            HelpView(self),
            self.view,
            align='center',
            width=('relative', 100),
            valign='middle',
            height=('relative', 100)
        ))

    def exit_help(self) -> None:
        self.loop.widget = self.view

    def search_messages(self, text: str) -> None:
        # Search for a text in messages
        self.update = False
        self.model.set_narrow(search=text)
        self.model.anchor = 10000000000
        self.model.num_after = 0
        self.model.num_before = 30
        self.model.get_messages(False)
        msg_id_list = self.model.index['search']
        w_list = create_msg_box_list(self.model, msg_id_list)
        self.model.msg_view.clear()
        self.model.msg_view.extend(w_list)
        focus_position = 0
        if focus_position >= 0 and focus_position < len(w_list):
            self.model.msg_list.set_focus(focus_position)

    def narrow_to_stream(self, button: Any) -> None:
        already_narrowed = self.model.set_narrow(stream=button.caption)
        if already_narrowed:
            return

        self.update = False
        # store the steam id in the model
        self.model.stream_id = button.stream_id
        # get the message ids of the current narrow
        msg_id_list = self.model.index['all_stream'][button.stream_id]
        # if no messages are found get more messages
        if len(msg_id_list) == 0:
            self.model.num_after = 10
            self.model.num_before = 30
            if hasattr(button, 'message'):
                self.model.anchor = button.message['id']
                self.model.get_messages(False)
            else:
                self.model.get_messages(True)
        msg_id_list = self.model.index['all_stream'][button.stream_id]
        if hasattr(button, 'message'):
            w_list = create_msg_box_list(
                self.model, msg_id_list, button.message['id'])
        else:
            w_list = create_msg_box_list(self.model, msg_id_list)

        self._finalize_show(w_list)

    def narrow_to_topic(self, button: Any) -> None:
        already_narrowed = self.model.set_narrow(stream=button.caption,
                                                 topic=button.title)
        if already_narrowed:
            return

        self.update = False
        self.model.stream_id = button.stream_id
        msg_id_list = self.model.index['stream'][button.stream_id].get(
                                                    button.title, [])
        if len(msg_id_list) == 0:
            first_anchor = True
            if hasattr(button, 'message'):
                self.model.anchor = button.message['id']
                first_anchor = False
            self.model.num_after = 10
            self.model.num_before = 30
            self.model.get_messages(first_anchor)
            msg_id_list = self.model.index['stream'][button.stream_id].get(
                                                    button.title, [])
        if hasattr(button, 'message'):
            w_list = create_msg_box_list(
                self.model, msg_id_list, button.message['id'])
        else:
            w_list = create_msg_box_list(self.model, msg_id_list)

        self._finalize_show(w_list)

    def narrow_to_user(self, button: Any) -> None:
        if hasattr(button, 'message'):
            emails = [recipient['email']
                      for recipient in button.message['display_recipient']
                      if recipient['email'] != self.model.client.email]
            user_emails = ', '.join(emails)
        else:
            user_emails = button.email

        already_narrowed = self.model.set_narrow(pm_with=user_emails)
        if already_narrowed:
            return

        button.user_id = self.model.user_dict[user_emails]['user_id']

        self.update = False
        msg_id_list = self.model.index['private'].get(frozenset(
            [self.model.user_id, button.user_id]), [])

        self.model.num_after = 10
        self.model.num_before = 30
        if hasattr(button, 'message'):
            self.model.anchor = button.message['id']
            self.model.get_messages(False)
        elif len(msg_id_list) == 0:
            self.model.get_messages(True)
        recipients = frozenset([self.model.user_id, button.user_id])
        self.model.recipients = recipients
        msg_id_list = self.model.index['private'].get(recipients, [])
        if hasattr(button, 'message'):
            w_list = create_msg_box_list(
                self.model, msg_id_list, button.message['id'])
        else:
            w_list = create_msg_box_list(self.model, msg_id_list)

        self._finalize_show(w_list)

    def show_all_messages(self, button: Any) -> None:
        already_narrowed = self.model.set_narrow()
        if already_narrowed:
            return

        self.update = False
        msg_list = self.model.index['all_messages']
        if hasattr(button, 'message'):
            w_list = create_msg_box_list(
                self.model, msg_list, button.message['id'])
        else:
            w_list = create_msg_box_list(self.model, msg_list)

        self._finalize_show(w_list)

    def show_all_pm(self, button: Any) -> None:
        already_narrowed = self.model.set_narrow(pm_with='')
        if already_narrowed:
            return

        self.update = False
        msg_list = self.model.index['all_private']
        if len(msg_list) == 0:
            self.model.num_after = 10
            self.model.num_before = 30
            self.model.get_messages(True)
            msg_list = self.model.index['all_private']
        w_list = create_msg_box_list(self.model, msg_list)

        self._finalize_show(w_list)

    def _finalize_show(self, w_list: List[Any]) -> None:
        focus_position = self.model.get_focus_in_current_narrow()

        if focus_position == set():
            focus_position = len(w_list) - 1
        assert not isinstance(focus_position, set)
        self.model.msg_view.clear()
        self.model.msg_view.extend(w_list)
        if focus_position >= 0 and focus_position < len(w_list):
            self.model.msg_list.set_focus(focus_position)

    @asyncn
    def register_initial_desired_events(self) -> None:
        event_types = [
            'message',
            'update_message',
            'reaction',
        ]
        response = self.client.register(event_types=event_types)
        self.max_message_id = response['max_message_id']
        self.queue_id = response['queue_id']
        self.last_event_id = response['last_event_id']

    def main(self) -> None:
        try:
            screen = Screen()
            screen.set_terminal_properties(colors=256)
            self.loop = urwid.MainLoop(self.view,
                                       self.view.palette[self.theme],
                                       screen=screen)
            self.update_pipe = self.loop.watch_pipe(self.draw_screen)
        except KeyError:
            print('Following are the themes available:')
            for theme in self.view.palette.keys():
                print(theme,)
            return

        try:
            self.loop.run()

        except Exception:
            self.restore_stdout()
            raise

        finally:
            self.restore_stdout()



from collections import defaultdict
from typing import Any, Dict

import pytest

from zulipterminal.ui_tools.boxes import MessageBox
from zulipterminal.ui_tools.buttons import StreamButton, UserButton


@pytest.fixture(autouse=True)
def no_requests(monkeypatch):
    """
    Forces all the tests to work offline.
    """
    monkeypatch.delattr("requests.sessions.Session.request")


@pytest.fixture(autouse=True)
def no_async(mocker):
    """
    Make all function calls synchronous.
    """
    mocker.patch('zulipterminal.helper.async')

# --------------- Controller Fixtures -----------------------------------------


@pytest.fixture
def stream_button(mocker):
    """
    Mocked stream button.
    """
    button = StreamButton(
        properties=['PTEST', 205, '#bfd56f'],
        controller=mocker.patch('zulipterminal.core.Controller'),
        view=mocker.patch('zulipterminal.ui.View')
    )
    return button


@pytest.fixture
def user_button(mocker):
    """
    Mocked User Button.
    """
    return UserButton(
        user={
            'user_id': 5179,
            'full_name': 'Boo Boo',
            'email': 'boo@zulip.com',
        },
        controller=mocker.patch('zulipterminal.core.Controller'),
        view=mocker.patch('zulipterminal.ui.View')
    )


@pytest.fixture
def msg_box(mocker, messages_successful_response):
    """
    Mocked MessageBox with stream message
    """
    return MessageBox(
        messages_successful_response['messages'][0],
        mocker.patch('zulipterminal.model.Model'),
        None,
    )


# --------------- Model Fixtures ----------------------------------------------


@pytest.fixture(scope='module')
def messages_successful_response() -> Dict[str, Any]:
    """
    A successful response from a /messages API query.
    """
    response = {
        'anchor': 10000000000000000,
        'messages': [{
            'id': 537286,
            'sender_full_name': 'Foo Foo',
            'timestamp': 1520918722,
            'client': 'website',
            'recipient_id': 6076,
            'sender_email': 'foo@zulip.com',
            'type': 'stream',
            'sender_realm_str': '',
            'flags': ['read'],
            'sender_id': 5140,
            'content_type': 'text/x-markdown',
            'stream_id': 205,
            'subject': 'Test',
            'reactions': [],
            'subject_links': [],
            'avatar_url': '/user_avatars/2/foo.png?x=x&version=2',
            'is_me_message': False,
            'sender_short_name': 'foo',
            'content': 'Stream content here.',
            'display_recipient': 'PTEST',
        }, {
            'id': 537287,
            'sender_full_name': 'Foo Foo',
            'timestamp': 1520918736,
            'client': 'website',
            'recipient_id': 5780,
            'is_me_message': False,
            'sender_email': 'foo@zulip.com',
            'flags': ['read'],
            'sender_id': 5140,
            'content_type': 'text/x-markdown',
            'sender_realm_str': '',
            'subject': '',
            'reactions': [],
            'type': 'private',
            'avatar_url': '/user_avatars/2/foo.png?x=x&version=2',
            'subject_links': [],
            'sender_short_name': 'foo',
            'content': 'Hey PM content here.',
            'display_recipient': [{
                'id': 5179,
                'is_mirror_dummy': False,
                'full_name': 'Boo Boo',
                'short_name': 'boo',
                'email': 'boo@zulip.com',
            }, {
                'short_name': 'foo',
                'id': 5140,
                'is_mirror_dummy': False,
                'full_name': 'Foo Foo',
                'email': 'foo@zulip.com',
            }],
        }],
        'result': 'success',
        'msg': '',
    }
    return response


@pytest.fixture(scope="module")
def initial_data():
    """
    Response from /register API request.
    """
    return {
        'unsubscribed': [{
            'audible_notifications': False,
            'description': 'announce',
            'stream_id': 7,
            'is_old_stream': True,
            'desktop_notifications': False,
            'pin_to_top': False,
            'stream_weekly_traffic': 0,
            'invite_only': False,
            'name': 'announce',
            'push_notifications': False,
            'email_address': '',
            'color': '#bfd56f',
            'in_home_view': True
        }],
        'result': 'success',
        'queue_id': '1522420755:786',
        'realm_users': [{
            'bot_type': None,
            'is_bot': False,
            'is_admin': False,
            'email': 'FOOBOO@gmail.com',
            'full_name': 'Tomás Farías',
            'user_id': 5827,
            'avatar_url': None,
            'is_active': True
        }, {
            'full_name': 'Jari Winberg',
            'user_id': 6086,
            'avatar_url': None,
            'is_active': True,
            'bot_type': None,
            'is_bot': False,
            'is_admin': False,
            'email': 'nyan.salmon+sns@gmail.com',
        }, {
            'bot_type': None,
            'is_bot': False,
            'is_admin': False,
            'email': 'cloudserver2@hotmail.de',
            'full_name': 'Test Account',
            'user_id': 6085,
            'is_active': True
        }],
        'subscriptions': [{
            'audible_notifications': False,
            'description': '',
            'stream_id': 86,
            'is_old_stream': True,
            'desktop_notifications': False,
            'pin_to_top': False,
            'stream_weekly_traffic': 0,
            'invite_only': False,
            'name': 'Django',
            'push_notifications': False,
            'email_address': '',
            'color': '#94c849',
            'in_home_view': True
        }, {
            'audible_notifications': False,
            'description': 'The Google Summer of Code',
            'stream_id': 14,
            'is_old_stream': True,
            'desktop_notifications': False,
            'pin_to_top': False,
            'stream_weekly_traffic': 53,
            'invite_only': False,
            'name': 'GSoC',
            'push_notifications': False,
            'email_address': '',
            'color': '#c2c2c2',
            'in_home_view': True
        }],
        'msg': '',
        'max_message_id': 552761,
        'never_subscribed': [{
            'invite_only': False,
            'description': 'Announcements from the Zulip GCI Mentors',
            'stream_id': 87,
            'name': 'GCI announce',
            'is_old_stream': True,
            'stream_weekly_traffic': 0
        }, {
            'invite_only': False,
            'description': 'General discussion',
            'stream_id': 74,
            'name': 'GCI general',
            'is_old_stream': True,
            'stream_weekly_traffic': 0
        }],
        'unread_msgs': {
            'pms': [],
            'count': 0,
            'mentions': [],
            'streams': [],
            'huddles': []
        },
        'presences': {
            'nyan.salmon+sns@gmail.com': {
                'ZulipElectron': {
                    'pushable': False,
                    'client': 'ZulipElectron',
                    'status': 'idle',
                    'timestamp': 1522484059
                },
                'ZulipMobile': {
                    'pushable': False,
                    'client': 'ZulipMobile',
                    'status': 'idle',
                    'timestamp': 1522384165
                },
                'aggregated': {
                    'timestamp': 1522484059,
                    'client': 'ZulipElectron',
                    'status': 'idle'
                }
            },
            'FOOBOO@gmail.com': {
                'website': {
                    'pushable': True,
                    'client': 'website',
                    'status': 'active',
                    'timestamp': 1522458138
                },
                'ZulipMobile': {
                    'pushable': True,
                    'client': 'ZulipMobile',
                    'status': 'active',
                    'timestamp': 1522480103
                },
                'aggregated': {
                    'timestamp': 1522480103,
                    'client': 'ZulipMobile',
                    'status': 'active'
                }
            }
        },
        'last_event_id': -1,
        'muted_topics': [],
    }


@pytest.fixture(scope="module")
def index_all_messages():
    """
    Expected index of `initial_data` fixture when model.narrow = []
    """
    return {
        'pointer': defaultdict(set, {}),
        'private': defaultdict(set, {}),
        'all_messages':  {537286, 537287},
        'all_private': set(),
        'messages': defaultdict(dict, {
            537286: {
                'type': 'stream',
                'sender_realm_str': '',
                'is_me_message': False,
                'content': 'Stream content here.',
                'recipient_id': 6076,
                'avatar_url': '/user_avatars/2/foo.png?x=x&version=2',
                'client': 'website',
                'stream_id': 205,
                'subject_links': [],
                'content_type': 'text/x-markdown',
                'display_recipient': 'PTEST',
                'reactions': [],
                'sender_short_name': 'foo',
                'id': 537286,
                'flags': ['read'],
                'sender_email': 'foo@zulip.com',
                'timestamp': 1520918722,
                'subject': 'Test',
                'sender_id': 5140,
                'sender_full_name': 'Foo Foo'
            },
            537287: {
                'type': 'private',
                'sender_realm_str': '',
                'is_me_message': False,
                'content': 'Hey PM content here.',
                'recipient_id': 5780,
                'client': 'website',
                'subject': '',
                'avatar_url': '/user_avatars/2/foo.png?x=x&version=2',
                'content_type': 'text/x-markdown',
                'display_recipient': [{
                    'id': 5179,
                    'full_name': 'Boo Boo',
                    'email': 'boo@zulip.com',
                    'short_name': 'boo',
                    'is_mirror_dummy': False
                }, {
                    'id': 5140,
                    'full_name': 'Foo Foo',
                    'email': 'foo@zulip.com',
                    'short_name': 'foo',
                    'is_mirror_dummy': False
                }],
                'sender_short_name': 'foo',
                'id': 537287,
                'flags': ['read'],
                'sender_email': 'foo@zulip.com',
                'timestamp': 1520918736,
                'reactions': [],
                'sender_id': 5140,
                'sender_full_name': 'Foo Foo',
                'subject_links': []
            }
        }),
        'all_stream': defaultdict(set, {}),
        'stream': defaultdict(dict, {}),
        'search': set(),
    }


@pytest.fixture(scope="module")
def index_stream():
    """
    Expected index of initial_data when model.narrow = [['stream', '7']]
    """
    return {
        'private': defaultdict(set, {}),
        'all_messages': set(),
        'all_private': {
            537287
        },
        'all_stream': defaultdict(set, {
            205: {
                537286
            }
        }),
        'search': set(),
        'stream': defaultdict(dict, {}),
        'pointer': defaultdict(set, {}),
        'messages': defaultdict(dict, {
            537286: {
                'is_me_message': False,
                'flags': ['read'],
                'content_type': 'text/x-markdown',
                'sender_realm_str': '',
                'timestamp': 1520918722,
                'type': 'stream',
                'sender_full_name': 'Foo Foo',
                'content': 'Stream content here.',
                'display_recipient': 'PTEST',
                'sender_id': 5140,
                'sender_email': 'foo@zulip.com',
                'sender_short_name': 'foo',
                'reactions': [],
                'client': 'website',
                'subject': 'Test',
                'avatar_url': '/user_avatars/2/foo.png?x=x&version=2',
                'recipient_id': 6076,
                'subject_links': [],
                'id': 537286,
                'stream_id': 205
            },
            537287: {
                'flags': ['read'],
                'content_type': 'text/x-markdown',
                'sender_realm_str': '',
                'timestamp': 1520918736,
                'type': 'private',
                'sender_full_name': 'Foo Foo',
                'content': 'Hey PM content here.',
                'display_recipient': [{
                    'email': 'boo@zulip.com',
                    'full_name': 'Boo Boo',
                    'short_name': 'boo',
                    'id': 5179,
                    'is_mirror_dummy': False
                }, {
                    'email': 'foo@zulip.com',
                    'full_name': 'Foo Foo',
                    'short_name': 'foo',
                    'id': 5140,
                    'is_mirror_dummy': False
                }],
                'sender_id': 5140,
                'sender_email': 'foo@zulip.com',
                'sender_short_name': 'foo',
                'reactions': [],
                'client': 'website',
                'subject': '',
                'avatar_url': '/user_avatars/2/foo.png?x=x&version=2',
                'recipient_id': 5780,
                'subject_links': [],
                'id': 537287,
                'is_me_message': False
            }
        })
    }


@pytest.fixture(scope="module")
def index_topic():
    """
    Expected index of initial_data when model.narrow = [['stream', '7'],
                                                        ['topic', 'Test']]
    """
    return {
        'stream': defaultdict(dict, {
            205: {
                'Test': {
                    537286
                }
            }
        }),
        'private': defaultdict(set, {}),
        'messages': defaultdict(dict, {
            537286: {
                'sender_full_name': 'Foo Foo',
                'sender_id': 5140,
                'id': 537286,
                'client': 'website',
                'is_me_message': False,
                'subject_links': [],
                'content': 'Stream content here.',
                'stream_id': 205,
                'sender_realm_str': '',
                'avatar_url': '/user_avatars/2/foo.png?x=x&version=2',
                'display_recipient': 'PTEST',
                'subject': 'Test',
                'sender_short_name': 'foo',
                'recipient_id': 6076,
                'content_type': 'text/x-markdown',
                'reactions': [],
                'timestamp': 1520918722,
                'flags': ['read'],
                'type': 'stream',
                'sender_email': 'foo@zulip.com'
            },
            537287: {
                'sender_full_name': 'Foo Foo',
                'sender_id': 5140,
                'id': 537287,
                'client': 'website',
                'avatar_url': '/user_avatars/2/foo.png?x=x&version=2',
                'subject_links': [],
                'content': 'Hey PM content here.',
                'sender_realm_str': '',
                'is_me_message': False,
                'display_recipient': [{
                    'is_mirror_dummy': False,
                    'id': 5179,
                    'full_name': 'Boo Boo',
                    'short_name': 'boo',
                    'email': 'boo@zulip.com'
                }, {
                    'short_name': 'foo',
                    'id': 5140,
                    'full_name': 'Foo Foo',
                    'is_mirror_dummy': False,
                    'email': 'foo@zulip.com'
                }],
                'subject': '',
                'sender_short_name': 'foo',
                'recipient_id': 5780,
                'content_type': 'text/x-markdown',
                'reactions': [],
                'timestamp': 1520918736,
                'flags': ['read'],
                'type': 'private',
                'sender_email': 'foo@zulip.com'
            }
        }),
        'pointer': defaultdict(set, {}),
        'all_private': set(),
        'all_stream': defaultdict(set, {}),
        'all_messages': set(),
        'search': set(),
    }


@pytest.fixture(scope="module")
def index_user():
    """
    Expected index of initial_data when model.narrow = [['pm_with',
                                                         'boo@zulip.com'],
    """
    return {
        'stream': defaultdict(dict, {}),
        'private': defaultdict(set, {
            frozenset({
                5179,
                5140
            }): {
                537287
            }
        }),
        'all_messages': set(),
        'pointer': defaultdict(set, {}),
        'messages': defaultdict(dict, {
            537286: {
                'subject': 'Test',
                'sender_full_name': 'Foo Foo',
                'sender_short_name': 'foo',
                'sender_email': 'foo@zulip.com',
                'is_me_message': False,
                'content_type': 'text/x-markdown',
                'type': 'stream',
                'id': 537286,
                'sender_id': 5140,
                'sender_realm_str': '',
                'stream_id': 205,
                'content': 'Stream content here.',
                'reactions': [],
                'subject_links': [],
                'client': 'website',
                'flags': ['read'],
                'timestamp': 1520918722,
                'avatar_url': '/user_avatars/2/foo.png?x=x&version=2',
                'recipient_id': 6076,
                'display_recipient': 'PTEST'
            },
            537287: {
                'subject': '',
                'sender_full_name': 'Foo Foo',
                'sender_short_name': 'foo',
                'sender_email': 'foo@zulip.com',
                'is_me_message': False,
                'content_type': 'text/x-markdown',
                'reactions': [],
                'id': 537287,
                'sender_id': 5140,
                'sender_realm_str': '',
                'type': 'private',
                'content': 'Hey PM content here.',
                'subject_links': [],
                'client': 'website',
                'flags': ['read'],
                'timestamp': 1520918736,
                'avatar_url': '/user_avatars/2/foo.png?x=x&version=2',
                'recipient_id': 5780,
                'display_recipient': [{
                    'short_name': 'boo',
                    'id': 5179,
                    'full_name': 'Boo Boo',
                    'is_mirror_dummy': False,
                    'email': 'boo@zulip.com'
                }, {
                    'id': 5140,
                    'short_name': 'foo',
                    'full_name': 'Foo Foo',
                    'is_mirror_dummy': False,
                    'email': 'foo@zulip.com'
                }]
            }
        }),
        'all_private': {
            537287
        },
        'all_stream': defaultdict(set, {}),
        'search': set(),
    }


@pytest.fixture(scope="module")
def user_profile():
    return {
        'max_message_id': 589270,
        'short_name': 'FOO',
        'full_name': 'FOO BOO',
        'email': 'FOO@ZULIP.COM',
        'is_bot': False,
        'user_id': 5140,
        'result': 'success',
        'client_id': 'abcd',
        'msg': '',
        'is_admin': False,
        'pointer': 589234
    }


@pytest.fixture(scope="module")
def error_response():
    return {
        "msg": "Invalid API key",
        "result": "error"
    }


@pytest.fixture(scope="module")
def user_dict():
    """
    User_dict created according to `initial_data` fixture.
    """
    return {
        'FOOBOO@gmail.com': {
            'full_name': 'Tomás Farías',
            'email': 'FOOBOO@gmail.com',
            'status': 'active',
            'user_id': 5827
        },
        'nyan.salmon+sns@gmail.com': {
            'full_name': 'Jari Winberg',
            'email': 'nyan.salmon+sns@gmail.com',
            'status': 'idle',
            'user_id': 6086
        },
        'cloudserver2@hotmail.de': {
            'full_name': 'Test Account',
            'email': 'cloudserver2@hotmail.de',
            'status': 'idle',
            'user_id': 6085
        }
    }


@pytest.fixture(scope="module")
def user_list():
    """
    List of users created corresponding to
    `initial_data` fixture.
    """
    return [{
        'full_name': 'Tomás Farías',
        'email': 'FOOBOO@gmail.com',
        'status': 'active',
        'user_id': 5827
    }, {
        'full_name': 'Jari Winberg',
        'email': 'nyan.salmon+sns@gmail.com',
        'status': 'idle',
        'user_id': 6086
    }, {
        'full_name': 'Test Account',
        'email': 'cloudserver2@hotmail.de',
        'status': 'idle',
        'user_id': 6085
    }]


@pytest.fixture(scope="module")
def streams():
    """
    List of streams created corresponding to
    `initial_data` fixture.
    """
    return [
        ['Django', 86, '#94c849'],
        ['GSoC', 14, '#c2c2c2']
    ]



import pytest
from zulipterminal.ui import View


class TestView:
    @pytest.fixture(autouse=True)
    def mock_external_classes(self, mocker):
        self.controller = mocker.patch('zulipterminal.core.Controller',
                                       return_value=None)
        self.client = mocker.patch('zulipterminal.core.Controller.client')
        self.model = mocker.patch('zulipterminal.core.Controller.model')
        self.write_box = mocker.patch('zulipterminal.ui.WriteBox')
        self.search_box = mocker.patch('zulipterminal.ui.SearchBox')

    @pytest.fixture
    def view(self, mocker):
        main_window = mocker.patch('zulipterminal.ui.View.main_window')
        return View(self.controller)

    def test_init(self, mocker):
        main_window = mocker.patch('zulipterminal.ui.View.main_window')
        view = View(self.controller)
        assert view.controller == self.controller
        assert view.model == self.model
        assert view.client == self.client
        assert view.streams == self.model.streams
        self.write_box.assert_called_once_with(view)
        self.search_box.assert_called_once_with(self.controller)
        main_window.assert_called_once_with()

    def test_left_column_view(self, mocker, view):
        left_view = mocker.patch('zulipterminal.ui.LeftColumnView')
        return_value = view.left_column_view()
        assert return_value == left_view(view)

    def test_message_view(self, view, mocker):
        middle_view = mocker.patch('zulipterminal.ui.MiddleColumnView')
        line_box = mocker.patch('zulipterminal.ui.urwid.LineBox')
        return_value = view.message_view()
        middle_view.assert_called_once_with(view.model, view.write_box,
                                            view.search_box)
        assert view.middle_column == middle_view()
        assert return_value == line_box()

    def test_right_column_view(self, view, mocker):
        right_view = mocker.patch('zulipterminal.ui.RightColumnView')
        line_box = mocker.patch('zulipterminal.ui.urwid.LineBox')
        return_value = view.right_column_view()
        right_view.assert_called_once_with(view)
        assert view.users_view == right_view()
        assert return_value == line_box()

    def test_main_window(self, mocker):
        left = mocker.patch('zulipterminal.ui.View.left_column_view')
        center = mocker.patch('zulipterminal.ui.View.message_view')
        right = mocker.patch('zulipterminal.ui.View.right_column_view')
        col = mocker.patch("zulipterminal.ui.urwid.Columns")
        frame = mocker.patch('zulipterminal.ui.urwid.Frame')
        title_divider = mocker.patch('zulipterminal.ui.urwid.Divider')
        text = mocker.patch('zulipterminal.ui.urwid.Text')

        full_name = "Bob James"
        email = "Bob@bob.com"
        server = "https://chat.zulip.zulip"

        mocker.patch('zulipterminal.core.Controller.client.get_profile',
                     return_value=dict(full_name=full_name, email=email))
        self.controller.client.base_url = server
        title_length = (len(email) + len(full_name) + len(server) + 9)

        view = View(self.controller)

        left.assert_called_once_with()
        center.assert_called_once_with()
        right.assert_called_once_with()
        expected_column_calls = [
            mocker.call([
                (25, left()),
                ('weight', 10, center()),
                (25, right()),
                ], focus_column=1),
            mocker.call([
                title_divider(),
                (title_length, text()),
                title_divider(),
                ])
        ]
        col.assert_has_calls(expected_column_calls)

        assert view.body == col()
        frame.assert_called_once_with(view.body, col(),
                                      focus_part='body')

    def test_keypress(self, view, mocker):
        view.users_view = mocker.Mock()
        view.body = mocker.Mock()
        view.user_search = mocker.Mock()
        size = (20,)
        key = 'j'

        super_view = mocker.patch("zulipterminal.ui.urwid.WidgetWrap.keypress")

        # Test Normal Mode keypress
        view.controller.editor_mode = False
        view.keypress(size, 'down')
        super_view.assert_called_once_with(size, 'down')

    def test_keypress_w(self, view, mocker):
        view.users_view = mocker.Mock()
        view.body = mocker.Mock()
        view.user_search = mocker.Mock()
        size = (20,)

        super_view = mocker.patch("zulipterminal.ui.urwid.WidgetWrap.keypress")
        view.controller.editor_mode = False

        # Test "w" keypress
        view.keypress(size, "w")
        view.users_view.keypress.assert_called_once_with(size, "w")
        assert view.body.focus_col == 2
        view.user_search.set_edit_text.assert_called_once_with("")
        assert view.controller.editor_mode is True
        assert view.controller.editor == view.user_search

    def test_keypress_q(self, view, mocker):
        view.stream_w = mocker.Mock()
        view.left_col_w = mocker.Mock()
        view.stream_w.search_box = mocker.Mock()
        view.body = mocker.Mock()
        size = (20,)

        super_view = mocker.patch("zulipterminal.ui.urwid.WidgetWrap.keypress")
        view.controller.editor_mode = False

        # Test "q" keypress
        view.keypress(size, "q")
        view.left_col_w.keypress.assert_called_once_with(size, "q")
        assert view.body.focus_col == 0
        view.stream_w.search_box.set_edit_text.assert_called_once_with("")
        assert view.controller.editor_mode is True
        assert view.controller.editor == view.stream_w.search_box

    def test_keypress_edit_mode(self, view, mocker):
        view.users_view = mocker.Mock()
        view.body = mocker.Mock()
        view.user_search = mocker.Mock()
        size = (20,)

        super_view = mocker.patch("zulipterminal.ui.urwid.WidgetWrap.keypress")

        # Test Edit Mode Keypress
        view.controller.editor_mode = True
        size = (130, 28)
        view.keypress(size, "w")
        view.controller.editor.keypress.assert_called_once_with((28,), "w")



from typing import Any, Dict, List, Tuple

import urwid

from zulipterminal.config import is_command_key


class MenuButton(urwid.Button):
    def __init__(self, caption: Any, email: str='') -> None:
        self.caption = caption  # str
        self.email = email
        super(MenuButton, self).__init__("")
        self._w = urwid.AttrMap(urwid.SelectableIcon(
            [self.caption], 0), None, 'selected')


class HomeButton(urwid.Button):
    def __init__(self, controller: Any, count: int=0) -> None:
        self.caption = 'All messages'
        self.count = count
        super(HomeButton, self).__init__("")
        self._w = self.widget(count)
        urwid.connect_signal(self, 'click', controller.show_all_messages)

    def update_count(self, count: int) -> None:
        self.count = count
        self._w = self.widget(count)

    def widget(self, count: int) -> Any:
        return urwid.AttrMap(urwid.SelectableIcon(
            [u' \N{BULLET} ', self.caption,
             ('idle', '' if count <= 0 else ' ' + str(count))],
            len(self.caption) + 4),
            None,
            'selected')


class PMButton(urwid.Button):
    def __init__(self, controller: Any, count: int=0) -> None:
        self.caption = 'Private messages'
        super(PMButton, self).__init__("")
        self.count = count
        self._w = self.widget(count)
        urwid.connect_signal(self, 'click', controller.show_all_pm)

    def update_count(self, count: int) -> None:
        self.count = count
        self._w = self.widget(count)

    def widget(self, count: int) -> Any:
        return urwid.AttrMap(urwid.SelectableIcon(
            [u' \N{BULLET} ', self.caption,
             ('idle', '' if count <= 0 else ' ' + str(count))],
            len(self.caption) + 4),
            None,
            'selected')


class StreamButton(urwid.Button):
    def __init__(self, properties: List[Any],
                 controller: Any, view: Any, count: int=0) -> None:
        self.caption = properties[0]
        self.stream_id = properties[1]
        color = properties[2]
        self.color = color[:2] + color[3] + color[5]
        view.palette['default'].append((self.color, '', '', '', self.color,
                                       'black'))
        view.palette['default'].append(('s' + self.color, '', '', '',
                                       'black', self.color))
        self.count = count
        super(StreamButton, self).__init__("")
        self._w = self.widget(count)
        urwid.connect_signal(self, 'click', controller.narrow_to_stream)

    def update_count(self, count: int) -> None:
        self.count = count
        self._w = self.widget(count)

    def widget(self, count: int) -> Any:
        return urwid.AttrMap(urwid.SelectableIcon(
            [(self.color, u' # '), self.caption,
             ('idle', '' if count <= 0 else ' ' + str(count))],
            len(self.caption) + 2),
            None,
            'selected')


class UserButton(urwid.Button):
    def __init__(self, user: Dict[str, Any], controller: Any,
                 view: Any, color: str=None, count: int=0) -> None:
        self.caption = user['full_name']  # str
        self.email = user['email']
        self.user_id = user['user_id']
        self.color = color
        self.count = count
        self.recipients = frozenset({
            self.user_id, view.model.user_id})
        super(UserButton, self).__init__("")
        self._w = self.widget(count)
        self.controller = controller
        self.view = view

    def update_count(self, count: int) -> None:
        self.count = count
        self._w = self.widget(count)

    def widget(self, count: int) -> Any:
        return urwid.AttrMap(urwid.SelectableIcon(
            [u'\N{BULLET} ', self.caption,
             ('idle', '' if count <= 0 else ' ' + str(count))],
            len(self.caption) + 2),
            self.color,
            'selected')

    def keypress(self, size: Tuple[int, int], key: str) -> str:
        if is_command_key('ENTER', key):
            self.controller.narrow_to_user(self)
            self.view.body.focus_col = 1
            self.view.body.focus.original_widget.set_focus('footer')
            self.view.write_box.private_box_view(self)
            return key
        return super(UserButton, self).keypress(size, key)


class TopicButton(urwid.Button):
    def __init__(self, stream_id: str, topic: str, model: Any) -> None:
        self.caption = model.stream_dict[int(stream_id)]['name']  # stream name
        self.title = topic
        self.stream_id = int(stream_id)


class UnreadPMButton(urwid.Button):
    def __init__(self, user_id: int, email: str) -> None:
        self.user_id = user_id
        self.email = email



from zulipterminal.helper import (
    update_flag,
    index_messages,
)
from typing import Any


def test_update_flag(mocker: Any) -> None:
    mock_controller = mocker.patch('zulipterminal.core.Controller')
    mock_api_query = mocker.patch('zulipterminal.core.Controller'
                                  '.client.do_api_query')
    update_flag([1, 2], mock_controller)
    mock_api_query.assert_called_once_with(
        {'flag': 'read', 'messages': [1, 2], 'op': 'add'},
        '/json/messages/flags',
        method='POST'
    )


def test_update_flag_empty_msg_list(mocker: Any) -> None:
    assert update_flag([], mocker.patch('zulip.Client')) is None


def test_index_messages_narrow_all_messages(mocker,
                                            messages_successful_response,
                                            index_all_messages) -> None:
    messages = messages_successful_response['messages']
    model = mocker.patch('zulipterminal.model.Model.__init__',
                         return_value=None)
    model.narrow = []
    assert index_messages(messages, model) == index_all_messages


def test_index_messages_narrow_stream(mocker,
                                      messages_successful_response,
                                      index_stream) -> None:
    messages = messages_successful_response['messages']
    model = mocker.patch('zulipterminal.model.Model.__init__',
                         return_value=None)
    model.narrow = [['stream', 'PTEST']]
    model.stream_id = 205
    assert index_messages(messages, model) == index_stream


def test_index_messages_narrow_topic(mocker,
                                     messages_successful_response,
                                     index_topic) -> None:
    messages = messages_successful_response['messages']
    model = mocker.patch('zulipterminal.model.Model.__init__',
                         return_value=None)
    model.narrow = [['stream', '7'],
                    ['topic', 'Test']]
    model.stream_id = 205
    assert index_messages(messages, model) == index_topic


def test_index_messages_narrow_user(mocker,
                                    messages_successful_response,
                                    index_user):
    messages = messages_successful_response['messages']
    model = mocker.patch('zulipterminal.model.Model.__init__',
                         return_value=None)
    model.narrow = [['pm_with', 'boo@zulip.com']]
    model.user_id = 5140
    model.user_dict = {
        'boo@zulip.com': {
            'user_id': 5179,
        }
    }
    assert index_messages(messages, model) == index_user



import argparse
import configparser
import traceback
import sys
import tempfile
from typing import Dict, Any
from os import path, remove

from zulipterminal.core import Controller


def parse_args() -> argparse.Namespace:
    description = '''
        Starts Zulip-Terminal.
        '''
    formatter_class = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(description=description,
                                     formatter_class=formatter_class)

    parser.add_argument('--config-file', '-c',
                        action='store',
                        help='config file downloaded from your zulip\
                             organization.(e.g. ~/zuliprc)')
    parser.add_argument('--theme', '-t',
                        default='default',
                        help='choose color theme. (e.g. blue, light)')
    # debug mode
    parser.add_argument("-d",
                        "--debug",
                        action="store_true",
                        help="Start zulip terminal in debug mode.")

    parser.add_argument('--profile', dest='profile',
                        action="store_true",
                        default=False, help='Profile runtime.')

    args = parser.parse_args()
    return args


def get_api_key(realm_url: str) -> Any:
    from getpass import getpass
    import requests

    email = input("\033[94mEmail:\033[0m ")
    password = getpass("\033[94mPassword:\033[0m ")
    response = requests.post(
        url=realm_url + '/api/v1/fetch_api_key',
        data={
            'username': email,
            'password': password,
        }
    )
    return response, email


def fetch_zuliprc(zuliprc_path: str) -> None:
    realm_url = input("\033[91mzuliprc file was not found"
                      " at " + zuliprc_path + "\033[0m"
                      "\nPlease enter your credentials to login into your"
                      " Zulip organization."
                      "\n\033[94mZulip-Realm URL:\033[0m ")
    if not realm_url.startswith('https://'):
        realm_url = 'https://' + realm_url
    res, email = get_api_key(realm_url)

    while res.status_code != 200:
        print("\n\033[91mUsername or Password Incorrect!\033[0m\n")
        res = get_api_key(realm_url)

    with open(zuliprc_path, 'w') as f:
        f.write('[api]' +
                '\nemail=' + email +
                '\nkey=' + str(res.json()['api_key']) +
                '\nsite=' + realm_url)
    print('Generated API key saved at ' + zuliprc_path)


def parse_zuliprc(zuliprc_str: str) -> Dict[str, Any]:
    zuliprc_path = path.expanduser(zuliprc_str)
    if not path.exists(zuliprc_path):
        try:
            fetch_zuliprc(zuliprc_path)
        except Exception:
            print('\n\033[91mInvalid Credentials, Please try again!\033[0m\n')
            # Remove zuliprc file if created.
            if path.exists(zuliprc_path):
                remove(zuliprc_path)
            parse_zuliprc(zuliprc_str)

    zuliprc = configparser.ConfigParser()
    zuliprc.read(zuliprc_path)

    # default settings
    settings = {'theme': 'default'}

    if 'zterm' in zuliprc:
        if 'theme' in zuliprc['zterm']:
            settings['theme'] = zuliprc['zterm']['theme']

    return settings


def main() -> None:
    """
    Launch Zulip Terminal.
    """

    args = parse_args()
    if args.config_file:
        zuliprc_path = args.config_file
    else:
        zuliprc_path = '~/zuliprc'

    zterm = parse_zuliprc(zuliprc_path)

    if args.profile:
        import cProfile
        prof = cProfile.Profile()
        prof.enable()

    try:
        Controller(zuliprc_path, zterm['theme']).main()
    except Exception as e:
        if args.debug:
            # A unexpected exception occurred, open the debugger in debug mode
            import pudb
            pudb.post_mortem()

        sys.stdout.flush()
        traceback.print_exc(file=sys.stderr)
        print("Zulip Terminal has crashed!", file=sys.stderr)
        print("You can ask for help at:", file=sys.stderr)
        print("https://chat.zulip.org/#narrow/stream/206-zulip-terminal",
              file=sys.stderr)
        print("\nThanks for using the Zulip-Terminal interface.\n")
        sys.stderr.flush()

    finally:
        if args.profile:
            prof.disable()
            prof.dump_stats("/tmp/profile.data")
            print("Profile data saved to /tmp/profile.data")
            print("You can visualize it using e.g."
                  "`snakeviz /tmp/profile.data`")

        sys.exit(1)


if __name__ == '__main__':
    main()



from collections import defaultdict
from time import ctime
from datetime import datetime
from typing import Any, Dict, List, Tuple, Union

import emoji
import urwid
from urwid_readline import ReadlineEdit

from zulipterminal.ui_tools.buttons import MenuButton
from zulipterminal.config import is_command_key


class WriteBox(urwid.Pile):
    def __init__(self, view: Any) -> None:
        super(WriteBox, self).__init__(self.main_view(True))
        self.client = view.client
        self.view = view

    def main_view(self, new: bool) -> Any:
        if new:
            return []
        else:
            self.contents.clear()

    def set_editor_mode(self) -> None:
        # if not in the editor mode already set editor_mode to True.
        if not self.view.controller.editor_mode:
            self.view.controller.editor_mode = True
            self.view.controller.editor = self

    def private_box_view(self, button: Any=None, email: str='') -> None:
        self.set_editor_mode()
        if email == '' and button is not None:
            email = button.email
        self.to_write_box = ReadlineEdit(u"To: ", edit_text=email)
        self.msg_write_box = ReadlineEdit(u"> ", multiline=True)
        self.contents = [
            (urwid.LineBox(self.to_write_box), self.options()),
            (self.msg_write_box, self.options()),
        ]
        self.focus_position = 1

    def stream_box_view(self, button: Any=None, caption: str='',
                        title: str='') -> None:
        self.set_editor_mode()
        self.to_write_box = None
        if caption == '' and button is not None:
            caption = button.caption
        self.msg_write_box = ReadlineEdit(u"> ", multiline=True)
        self.stream_write_box = ReadlineEdit(
            caption=u"Stream:  ",
            edit_text=caption
        )
        self.title_write_box = ReadlineEdit(caption=u"Title:  ",
                                            edit_text=title)

        header_write_box = urwid.Columns([
            urwid.LineBox(self.stream_write_box),
            urwid.LineBox(self.title_write_box),
        ])
        write_box = [
            (header_write_box, self.options()),
            (self.msg_write_box, self.options()),
        ]
        self.contents = write_box

    def keypress(self, size: Tuple[int, int], key: str) -> str:
        if is_command_key('SEND_MESSAGE', key):
            if not self.to_write_box:
                request = {
                    'type': 'stream',
                    'to': self.stream_write_box.edit_text,
                    'subject': self.title_write_box.edit_text,
                    'content': self.msg_write_box.edit_text,
                }
                response = self.client.send_message(request)
            else:
                request = {
                    'type': 'private',
                    'to': self.to_write_box.edit_text,
                    'content': self.msg_write_box.edit_text,
                }
                response = self.client.send_message(request)
            if response['result'] == 'success':
                self.msg_write_box.edit_text = ''
        elif is_command_key('GO_BACK', key):
            self.view.controller.editor_mode = False
            self.main_view(False)
            self.view.middle_column.set_focus('body')
        elif is_command_key('GO_RIGHT', key) and self.to_write_box is None:
            self.contents[0][0].focus_col = 1
        elif is_command_key('GO_LEFT', key) and self.to_write_box is None:
            self.contents[0][0].focus_col = 0
        elif is_command_key('TAB', key):
            if len(self.contents) == 0:
                return key
            # toggle focus position
            if self.focus_position == 0 and self.to_write_box is None:
                if self.contents[0][0].focus_col == 0:
                    self.contents[0][0].focus_col = 1
                    return key
                else:
                    self.contents[0][0].focus_col = 0
            self.focus_position = self.focus_position == 0
            self.contents[0][0].focus_col = 0

        key = super(WriteBox, self).keypress(size, key)
        return key


class MessageBox(urwid.Pile):
    def __init__(self, message: Dict[str, Any], model: Any,
                 last_message: Any) -> None:
        self.model = model
        self.message = message
        self.caption = ''
        self.stream_id = None  # type: Union[int, None]
        self.title = ''
        self.email = ''
        self.last_message = last_message
        # if this is the first message
        if self.last_message is None:
            self.last_message = defaultdict(dict)
        super(MessageBox, self).__init__(self.main_view())

    def _time_for_message(self, message: Dict[str, Any]) -> str:
        return ctime(message['timestamp'])[:-8]

    def stream_view(self) -> Any:
        self.caption = self.message['display_recipient']
        self.stream_id = self.message['stream_id']
        self.title = self.message['subject']
        # If the topic of last message is same
        # as current message
        if self.title == self.last_message['subject'] and\
                self.last_message['type'] == 'stream':
            return None
        bar_color = self.model.stream_dict[self.stream_id]['color']
        bar_color = 's' + bar_color[:2] + bar_color[3] + bar_color[5]
        stream_title_markup = ('bar', [
            (bar_color, self.caption),
            (None, '>'),
            ('title', self.title)
        ])
        stream_title = urwid.Text(stream_title_markup)
        header = urwid.AttrWrap(stream_title, 'bar')
        header.markup = stream_title_markup
        return header

    def private_view(self) -> Any:
        self.email = self.message['sender_email']
        self.user_id = self.message['sender_id']

        recipient_ids = [{recipient['id']
                          for recipient in message['display_recipient']
                          if 'id' in recipient}
                         for message in (self.message, self.last_message)
                         if 'display_recipient' in message]
        if len(recipient_ids) == 2 and\
                recipient_ids[0] == recipient_ids[1] and\
                self.last_message['type'] == 'private':
            return None
        self.recipients = ', '.join(list(
            recipient['full_name']
            for recipient in self.message['display_recipient']
            if recipient['email'] != self.model.client.email
        ))
        title_markup = ('header', [
            ('custom', 'Private Messages with'),
            ('selected', ": "),
            ('custom', self.recipients)
        ])
        title = urwid.Text(title_markup)
        header = urwid.AttrWrap(title, 'bar')
        header.markup = title_markup
        return header

    def reactions_view(self, reactions: List[Dict[str, Any]]) -> Any:
        if not reactions:
            return ''
        try:
            reacts = defaultdict(int)  # type: Dict[str, int]
            custom_reacts = defaultdict(int)  # type: Dict[str, int]
            for reaction in reactions:
                if reaction['reaction_type'] == 'unicode_emoji':
                    reacts[reaction['emoji_code']] += 1
                elif reaction['reaction_type'] == 'realm_emoji':
                    custom_reacts[reaction['emoji_name']] += 1
            dis = [
                '\\U' + '0'*(8-len(emoji)) + emoji + ' ' + str(reacts[emoji]) +
                ' ' for emoji in reacts]
            emojis = ''.join(e.encode().decode('unicode-escape') for e in dis)
            custom_emojis = ''.join(
                ['{} {}'.format(r, custom_reacts[r]) for r in custom_reacts])
            return urwid.Padding(
                urwid.Text(([
                    ('emoji', emoji.demojize(emojis + custom_emojis))
                ])), align='left', width=('relative', 50), left=35,
                min_width=50)
        except Exception:
            return ''

    def main_view(self) -> List[Any]:
        if self.message['type'] == 'stream':
            header = self.stream_view()
        else:
            header = self.private_view()

        reactions = self.reactions_view(self.message['reactions'])

        content = [emoji.demojize(self.message['content'])]
        content = urwid.Padding(urwid.Text(content),
                                align='left', width=('relative', 50), left=35,
                                min_width=50)

        message_author = self.message['sender_full_name']
        message_time = self._time_for_message(self.message)

        # Statements as to how the message varies from the previous one
        different_topic = header is not None
        different_author = (
            self.last_message['sender_full_name'] != message_author)
        more_than_24h_apart = (
            'timestamp' in self.last_message and
            (datetime.fromtimestamp(self.message['timestamp']) -
             datetime.fromtimestamp(self.last_message['timestamp'])).days)
        different_timestamp = (
            'timestamp' in self.last_message and
            message_time != self._time_for_message(self.last_message))

        # Include author name/time under various conditions
        author_time_items = []
        if different_topic or different_author or more_than_24h_apart:
            author_time_items.append(urwid.Text([('name', message_author)]))
        if different_topic or different_author or different_timestamp:
            author_time_items.append(urwid.Text([('time', message_time)],
                                                align='right'))
        author_and_time = urwid.Columns(author_time_items)

        view = [header, author_and_time, content, reactions]
        if header is None:
            view.remove(header)
        if not author_time_items:
            view.remove(author_and_time)
        if reactions == '':
            view.remove(reactions)
        return view

    def selectable(self) -> bool:
        return True

    def mouse_event(self, size: Tuple[int, int], event: Any, button: Any,
                    col: int, row: int, focus: int) -> Union[bool, Any]:
        if event == 'mouse press':
            if button == 1:
                self.keypress(size, 'enter')
                return True
        return super(MessageBox, self).mouse_event(size, event, button, col,
                                                   row, focus)

    def get_recipients(self) -> str:
        emails = []
        for recipient in self.message['display_recipient']:
            email = recipient['email']
            if email == self.model.client.email:
                continue
            emails.append(recipient['email'])
        return ', '.join(emails)

    def keypress(self, size: Tuple[int, int], key: str) -> str:
        if is_command_key('ENTER', key):
            if self.message['type'] == 'private':
                self.model.controller.view.write_box.private_box_view(
                    email=self.get_recipients()
                )
            elif self.message['type'] == 'stream':
                self.model.controller.view.write_box.stream_box_view(
                    caption=self.message['display_recipient'],
                    title=self.message['subject']
                )
        elif is_command_key('STREAM_MESSAGE', key):
            if self.message['type'] == 'private':
                self.model.controller.view.write_box.private_box_view(
                    email=self.get_recipients()
                )
            elif self.message['type'] == 'stream':
                self.model.controller.view.write_box.stream_box_view(
                    caption=self.message['display_recipient']
                )
        elif is_command_key('STREAM_NARROW', key):
            if self.message['type'] == 'private':
                self.model.controller.narrow_to_user(self)
            elif self.message['type'] == 'stream':
                self.model.controller.narrow_to_stream(self)
        elif is_command_key('TOPIC_NARROW', key):
            if self.message['type'] == 'private':
                self.model.controller.narrow_to_user(self)
            elif self.message['type'] == 'stream':
                self.model.controller.narrow_to_topic(self)
        elif is_command_key('GO_BACK', key):
            self.model.controller.show_all_messages(self)
        elif is_command_key('REPLY_AUTHOR', key):
            self.model.controller.view.write_box.private_box_view(
                email=self.message['sender_email']
            )
        elif is_command_key('ALL_PM', key):
            self.model.controller.show_all_pm(self)
        return key


class SearchBox(urwid.Pile):
    def __init__(self, controller: Any) -> None:
        self.controller = controller
        super(SearchBox, self).__init__(self.main_view())

    def main_view(self) -> Any:
        self.text_box = ReadlineEdit(u"Search: ")
        # Add some text so that when packing,
        # urwid doesn't hide the widget.
        self.msg_narrow = urwid.Text("DONT HIDE")
        w = urwid.Columns([
            ('pack', self.msg_narrow),
            ('pack', urwid.Text("  ")),
            self.text_box,
        ])
        self.w = urwid.LineBox(
            w, tlcorner=u'', tline=u'', lline=u'',
            trcorner=u'', blcorner=u'─', rline=u'',
            bline=u'─', brcorner=u'─')
        return [self.w]

    def keypress(self, size: Tuple[int, int], key: str) -> str:
        if is_command_key('GO_BACK', key):
            self.text_box.set_edit_text("")
            self.controller.editor_mode = False
            self.controller.view.middle_column.set_focus('body')
            return key

        elif is_command_key('ENTER', key):
            self.controller.editor_mode = False
            self.controller.model.index['search'] = set()
            self.controller.search_messages(self.text_box.edit_text)
            self.controller.view.middle_column.set_focus('body')
            return key

        key = super(SearchBox, self).keypress(size, key)
        return key


class UserSearchBox(urwid.Edit):
    """
    Search Box to search users in real-time.
    """

    def __init__(self, user_view: Any) -> None:
        self.user_view = user_view
        super(UserSearchBox, self).__init__(edit_text="Search people")

    def keypress(self, size: Tuple[int, int], key: str) -> str:
        if is_command_key('ENTER', key):
            self.user_view.view.controller.editor_mode = False
            self.user_view.set_focus("body")
        if is_command_key('GO_BACK', key):
            self.user_view.view.controller.editor_mode = False
            self.set_edit_text("Search people")
            self.user_view.set_focus("body")
            self.user_view.keypress(size, 'esc')

        return super(UserSearchBox, self).keypress(size, key)


class StreamSearchBox(urwid.Edit):
    """
    Search Box to search streams in real-time.urwid
    """

    def __init__(self, stream_view: Any) -> None:
        self.stream_view = stream_view
        super(StreamSearchBox, self).__init__(edit_text="Search streams")

    def keypress(self, size: Tuple[int, int], key: str) -> str:
        if is_command_key('ENTER', key):
            self.stream_view.view.controller.editor_mode = False
            self.stream_view.set_focus("body")
            self.stream_view.body.set_focus(0)
        if is_command_key('GO_BACK', key):
            self.stream_view.view.controller.editor_mode = False
            self.set_edit_text("Search streams")
            self.stream_view.set_focus("body")
            self.stream_view.keypress(size, 'esc')

        return super(StreamSearchBox, self).keypress(size, key)



from platform import platform
from typing import Any

import pytest

from zulipterminal.core import Controller


class TestController:
    @pytest.fixture(autouse=True)
    def mock_external_classes(self, mocker: Any) -> None:
        self.client = mocker.patch('zulip.Client')
        self.model = mocker.patch('zulipterminal.model.Model.__init__',
                                  return_value=None)
        self.view = mocker.patch('zulipterminal.ui.View.__init__',
                                 return_value=None)
        self.model.poll_for_events = mocker.patch('zulipterminal.model.Model'
                                                  '.poll_for_events')
        mocker.patch('zulipterminal.core.Controller.show_loading')

    @pytest.fixture
    def controller(self, mocker) -> None:
        self.config_file = 'path/to/zuliprc'
        self.theme = 'default'
        mocker.patch('zulipterminal.core.Controller.'
                     'register_initial_desired_events')
        return Controller(self.config_file, self.theme)

    def test_initialize_controller(self, controller, mocker) -> None:
        self.client.assert_called_once_with(
            config_file=self.config_file,
            client='ZulipTerminal/0.1.0 ' + platform(),
        )
        self.model.assert_called_once_with(controller)
        self.view.assert_called_once_with(controller)
        self.model.poll_for_events.assert_called_once_with()
        controller.register_initial_desired_events.assert_called_once_with()
        assert controller.theme == self.theme

    def test_narrow_to_stream(self, mocker, controller,
                              stream_button, index_stream) -> None:
        controller.model.narrow = []
        controller.model.index = index_stream
        controller.model.msg_view = mocker.patch('urwid.SimpleFocusListWalker')
        controller.model.msg_list = mocker.patch('urwid.ListBox')
        controller.model.stream_dict = {
            205: {
                'color': '#ffffff',
            }
        }
        controller.model.muted_streams = []
        controller.model.muted_topics = []
        controller.narrow_to_stream(stream_button)
        assert controller.model.stream_id == stream_button.stream_id
        assert controller.model.narrow == [['stream', stream_button.caption]]
        controller.model.msg_view.clear.assert_called_once_with()
        controller.model.msg_list.set_focus.assert_called_once_with(0)
        widget = controller.model.msg_view.extend.call_args_list[0][0][0][0]
        id_list = index_stream['all_stream'][stream_button.stream_id]
        assert {widget.original_widget.message['id']} == id_list

    def test_narrow_to_topic(self, mocker, controller,
                             msg_box, index_topic):
        expected_narrow = [['stream', msg_box.caption],
                           ['topic', msg_box.title]]
        controller.model.narrow = []
        controller.model.index = index_topic
        controller.model.msg_view = mocker.patch('urwid.SimpleFocusListWalker')
        controller.model.msg_list = mocker.patch('urwid.ListBox')
        controller.model.stream_dict = {
            205: {
                'color': '#ffffff',
            }
        }
        controller.model.muted_streams = []
        controller.model.muted_topics = []
        controller.narrow_to_topic(msg_box)
        assert controller.model.stream_id == msg_box.stream_id
        assert controller.model.narrow == expected_narrow
        controller.model.msg_view.clear.assert_called_once_with()
        controller.model.msg_list.set_focus.assert_called_once_with(0)
        widget = controller.model.msg_view.extend.call_args_list[0][0][0][0]
        id_list = index_topic['stream'][msg_box.stream_id][msg_box.title]
        assert {widget.original_widget.message['id']} == id_list

    def test_narrow_to_user(self, mocker, controller, user_button, index_user):
        controller.model.client = self.client
        controller.model.narrow = []
        controller.model.index = index_user
        controller.model.msg_view = mocker.patch('urwid.SimpleFocusListWalker')
        controller.model.msg_list = mocker.patch('urwid.ListBox')
        controller.model.user_id = 5140
        controller.model.user_dict = {
            user_button.email: {
                'user_id': user_button.user_id
            }
        }
        controller.narrow_to_user(user_button)
        assert controller.model.narrow == [["pm_with", user_button.email]]
        controller.model.msg_view.clear.assert_called_once_with()
        controller.model.msg_list.set_focus.assert_called_once_with(0)
        recipients = frozenset([controller.model.user_id, user_button.user_id])
        assert controller.model.recipients == recipients
        widget = controller.model.msg_view.extend.call_args_list[0][0][0][0]
        id_list = index_user['private'][recipients]
        assert {widget.original_widget.message['id']} == id_list

    def test_show_all_messages(self, mocker, controller, index_all_messages):
        controller.model.client = self.client
        controller.model.narrow = [['stream', 'PTEST']]
        controller.model.index = index_all_messages
        controller.model.msg_view = mocker.patch('urwid.SimpleFocusListWalker')
        controller.model.msg_list = mocker.patch('urwid.ListBox')
        controller.model.stream_dict = {
            205: {
                'color': '#ffffff',
            }
        }
        controller.model.muted_streams = []
        controller.model.muted_topics = []
        controller.show_all_messages('')
        assert controller.model.narrow == []
        controller.model.msg_view.clear.assert_called_once_with()
        controller.model.msg_list.set_focus.assert_called_once_with(1)
        widgets = controller.model.msg_view.extend.call_args_list[0][0][0]
        id_list = index_all_messages['all_messages']
        msg_ids = set(widget.original_widget.message['id']
                      for widget in widgets)
        assert msg_ids == id_list

    def test_show_all_pm(self, mocker, controller, index_user):
        controller.model.client = self.client
        controller.model.narrow = []
        controller.model.index = index_user
        controller.model.msg_view = mocker.patch('urwid.SimpleFocusListWalker')
        controller.model.msg_list = mocker.patch('urwid.ListBox')
        controller.show_all_pm('')
        assert controller.model.narrow == [['is', 'private']]
        controller.model.msg_view.clear.assert_called_once_with()
        controller.model.msg_list.set_focus.assert_called_once_with(0)
        id_list = index_user['all_private']
        widget = controller.model.msg_view.extend.call_args_list[0][0][0][0]
        assert {widget.original_widget.message['id']} == id_list

    def test_register_initial_desired_events(self, mocker):
        self.config_file = 'path/to/zuliprc'
        self.theme = 'default'
        controller = Controller(self.config_file, self.theme)
        event_types = [
            'message',
            'update_message',
            'reaction'
        ]
        controller.client.register.assert_called_once_with(
                                   event_types=event_types)

    def test_main(self, mocker, controller):
        ret_mock = mocker.Mock()
        mock_loop = mocker.patch('urwid.MainLoop', return_value=ret_mock)
        controller.view.palette = {
            'default': 'theme_properties'
        }
        controller.main()
        assert mock_loop.call_count == 1



from collections import defaultdict
from typing import Any, List, Tuple
import threading

import urwid

from zulipterminal.config import KEY_BINDINGS, is_command_key
from zulipterminal.helper import asyncn, update_flag, match_user
from zulipterminal.ui_tools.buttons import (
    TopicButton,
    UnreadPMButton,
    UserButton,
    HomeButton,
    PMButton,
    StreamButton,
)
from zulipterminal.ui_tools.utils import create_msg_box_list
from zulipterminal.ui_tools.boxes import UserSearchBox, StreamSearchBox


class MessageView(urwid.ListBox):
    def __init__(self, model: Any) -> None:
        self.model = model
        self.index = model.index
        # Initialize for reference
        self.focus_msg = 0
        self.log = urwid.SimpleFocusListWalker(self.main_view())
        urwid.connect_signal(self.log, 'modified', self.read_message)
        # This Function completely controls the messages
        # shown in the MessageView
        self.model.msg_view = self.log

        super(MessageView, self).__init__(self.log)
        self.set_focus(self.focus_msg)
        # if loading new/old messages - True
        self.old_loading = False
        self.new_loading = False

    def main_view(self) -> List[Any]:
        msg_btn_list = create_msg_box_list(self.model)
        focus_msg = self.model.get_focus_in_current_narrow()
        if focus_msg == set():
            focus_msg = len(msg_btn_list) - 1
        self.focus_msg = focus_msg
        return msg_btn_list

    @asyncn
    def load_old_messages(self, anchor: int=10000000000) -> None:
        self.old_loading = True
        # Use the currently focused image as anchor
        self.model.anchor = anchor
        # We don't want message after the current message
        self.model.num_after = 0
        self.model.num_before = 30
        current_ids = self.model.get_message_ids_in_current_narrow()
        self.index = self.model.get_messages(False)
        msg_ids = self.model.get_message_ids_in_current_narrow() - current_ids
        message_list = create_msg_box_list(self.model, msg_ids)
        message_list.reverse()
        for msg_w in message_list:
            self.log.insert(0, msg_w)
        self.model.controller.update_screen()
        self.old_loading = False

    @asyncn
    def load_new_messages(self, anchor: int) -> None:
        self.new_loading = True
        self.model.anchor = anchor
        self.model.num_before = 0
        self.model.num_after = 30
        current_ids = self.model.get_message_ids_in_current_narrow()
        self.index = self.model.get_messages(False)
        msg_ids = self.model.get_message_ids_in_current_narrow() - current_ids
        message_list = create_msg_box_list(self.model, msg_ids)
        self.log.extend(message_list)
        self.model.controller.update_screen()
        self.new_loading = False

    def mouse_event(self, size: Any, event: str, button: int, col: int,
                    row: int, focus: Any) -> Any:
        if event == 'mouse press':
            if button == 4:
                self.keypress(size, 'up')
                return True
            if button == 5:
                self.keypress(size, 'down')
                return True
        return super(MessageView, self).mouse_event(size, event, button, col,
                                                    row, focus)

    def keypress(self, size: Tuple[int, int], key: str) -> str:
        if is_command_key('NEXT_MESSAGE', key) and not self.new_loading:
            try:
                position = self.log.next_position(self.focus_position)
                self.set_focus(position, 'above')
                return key
            except Exception:
                if self.focus:
                    id = self.focus.original_widget.message['id']
                    self.load_new_messages(id)
                return key

        elif is_command_key('PREVIOUS_MESSAGE', key) and not self.old_loading:
            try:
                position = self.log.prev_position(self.focus_position)
                self.set_focus(position, 'below')
                return key
            except Exception:
                if self.focus:
                    id = self.focus.original_widget.message['id']
                    self.load_old_messages(id)
                else:
                    self.load_old_messages()
                return key

        elif is_command_key('SCROLL_TO_TOP', key) and not self.old_loading:
            if self.focus_position == 0:
                return self.keypress(size, 'up')
            else:
                return super(MessageView, self).keypress(size, 'page up')

        elif is_command_key('SCROLL_TO_BOTTOM', key) and not self.old_loading:
            if self.focus_position == len(self.log) - 1:
                return self.keypress(size, 'down')
            else:
                return super(MessageView, self).keypress(size, 'page down')

        key = super(MessageView, self).keypress(size, key)
        return key

    def update_current_footer(self, message_view: Any) -> None:
        if not hasattr(self.model.controller, 'view'):
            return
        # if view is ready display current narrow
        # at the bottom of the view.
        message_view.last_message = defaultdict(dict)
        is_stream = message_view.message['type'] == 'stream'
        if is_stream:
            footer = message_view.stream_view()
        else:
            footer = message_view.private_view()
        self.model.controller.view.search_box.msg_narrow.set_text(
            footer.markup
        )
        self.model.controller.update_screen()

    def read_message(self) -> None:
        # Message currently in focus
        msg_w, curr_pos = self.body.get_focus()
        if msg_w is None:
            return
        self.update_current_footer(msg_w.original_widget)
        # save the current focus
        self.model.set_focus_in_current_narrow(self.focus_position)
        # msg ids that have been read
        read_msg_ids = list()  # type: List[int]
        # until we find a read message above the current message
        while msg_w.attr_map == {None: 'unread'}:
            msg_id = msg_w.original_widget.message['id']
            read_msg_ids.append(msg_id)
            self.model.index['messages'][msg_id]['flags'].append('read')
            msg_w.set_attr_map({None: None})
            msg_w, curr_pos = self.body.get_prev(curr_pos)
            if msg_w is None:
                break
        update_flag(read_msg_ids, self.model.controller)


class StreamsView(urwid.Frame):
    def __init__(self, streams_btn_list: List[Any], view: Any) -> None:
        self.view = view
        self.log = urwid.SimpleFocusListWalker(streams_btn_list)
        self.streams_btn_list = streams_btn_list
        list_box = urwid.ListBox(self.log)
        self.search_box = StreamSearchBox(self)
        urwid.connect_signal(self.search_box, 'change', self.update_streams)
        super(StreamsView, self).__init__(list_box, header=urwid.LineBox(
            self.search_box, tlcorner=u'─', tline=u'─', lline=u'',
            trcorner=u'─', blcorner=u'─', rline=u'',
            bline=u'─', brcorner=u'─'
        ))
        self.search_lock = threading.Lock()

    @asyncn
    def update_streams(self, search_box: Any, new_text: str) -> None:
        if not self.view.controller.editor_mode:
            return
        # wait for any previously started search to finish to avoid
        # displaying wrong stream list.
        self.search_lock.acquire()
        streams_display = self.streams_btn_list.copy()
        for stream in self.streams_btn_list:
            if not stream.caption.lower().startswith(new_text):
                streams_display.remove(stream)
        self.log.clear()
        self.log.extend(streams_display)
        self.view.controller.update_screen()
        self.search_lock.release()

    def mouse_event(self, size: Any, event: str, button: int, col: int,
                    row: int, focus: Any) -> Any:
        if event == 'mouse press':
            if button == 4:
                self.keypress(size, 'up')
                return True
            elif button == 5:
                self.keypress(size, 'down')
                return True
        return super(StreamsView, self).mouse_event(size, event, button, col,
                                                    row, focus)

    def keypress(self, size: Tuple[int, int], key: str) -> str:
        if is_command_key('SEARCH_STREAMS', key):
            self.set_focus('header')
            return key
        elif is_command_key('GO_BACK', key):
            self.search_box.set_edit_text("Search streams")
            self.log.clear()
            self.log.extend(self.streams_btn_list)
            self.set_focus('body')
            self.view.controller.update_screen()
            return key
        return super(StreamsView, self).keypress(size, key)


class UsersView(urwid.ListBox):
    def __init__(self, users_btn_list: List[Any]) -> None:
        self.log = urwid.SimpleFocusListWalker(users_btn_list)
        super(UsersView, self).__init__(self.log)

    def mouse_event(self, size: Any, event: str, button: int, col: int,
                    row: int, focus: Any) -> Any:
        if event == 'mouse press':
            if button == 4:
                for _ in range(5):
                    self.keypress(size, 'up')
                return True
            elif button == 5:
                for _ in range(5):
                    self.keypress(size, 'down')
        return super(UsersView, self).mouse_event(size, event, button, col,
                                                  row, focus)


class MiddleColumnView(urwid.Frame):
    def __init__(self, model: Any, write_box: Any, search_box: Any) -> None:
        msg_list = MessageView(model)
        self.model = model
        self.controller = model.controller
        self.last_unread_topic = None
        self.last_unread_pm = None
        self.search_box = search_box
        model.msg_list = msg_list
        super(MiddleColumnView, self).__init__(msg_list, header=search_box,
                                               footer=write_box)

    def get_next_unread_topic(self) -> Any:
        topics = list(self.model.unread_counts['unread_topics'].keys())
        next_topic = False
        for topic in topics:
            if next_topic is True:
                self.last_unread_topic = topic
                return topic
            if topic == self.last_unread_topic:
                next_topic = True
        if len(topics) > 0:
            topic = topics[0]
            self.last_unread_topic = topic
            return topic
        return

    def get_next_unread_pm(self) -> Any:
        pms = list(self.model.unread_counts['unread_pms'].keys())
        next_pm = False
        for pm in pms:
            if next_pm is True:
                self.last_unread_pm = pm
                return pm
            if pm == self.last_unread_pm:
                next_pm = True
        if len(pms) > 0:
            pm = pms[0]
            self.last_unread_pm = pm
            return pm
        return

    def keypress(self, size: Tuple[int, int], key: str) -> str:
        if is_command_key('GO_BACK', key):
            self.header.keypress(size, 'esc')
            self.footer.keypress(size, 'esc')
            self.set_focus('body')

        elif self.focus_position in ['footer', 'header']:
            return super(MiddleColumnView, self).keypress(size, key)

        elif is_command_key('SEARCH_MESSAGES', key):
            self.controller.editor_mode = True
            self.controller.editor = self.search_box
            self.set_focus('header')
            return key

        elif is_command_key('REPLY_MESSAGE', key):
            self.body.keypress(size, 'enter')
            self.set_focus('footer')
            self.footer.focus_position = 1
            return key

        elif is_command_key('STREAM_MESSAGE', key):
            self.body.keypress(size, 'c')
            self.set_focus('footer')
            self.footer.focus_position = 0
            return key

        elif is_command_key('REPLY_AUTHOR', key):
            self.body.keypress(size, 'R')
            self.set_focus('footer')
            self.footer.focus_position = 1
            return key

        elif is_command_key('NEXT_UNREAD_TOPIC', key):
            # narrow to next unread topic
            stream_topic = self.get_next_unread_topic()
            if stream_topic is None:
                return key
            stream, topic = stream_topic
            self.controller.narrow_to_topic(TopicButton(stream, topic,
                                                        self.model))
            return key
        elif is_command_key('NEXT_UNREAD_PM', key):
            # narrow to next unread pm
            pm = self.get_next_unread_pm()
            if pm is None:
                return key
            email = self.model.user_id_email_dict[pm]
            self.controller.narrow_to_user(UnreadPMButton(pm, email))
        elif is_command_key('PRIVATE_MESSAGE', key):
            # Create new PM message
            self.footer.private_box_view()
            self.set_focus('footer')
            self.footer.focus_position = 0
            return key
        return super(MiddleColumnView, self).keypress(size, key)


class RightColumnView(urwid.Frame):
    """
    Displays the users list on the right side of the app.
    """

    def __init__(self, view: Any) -> None:
        self.view = view
        self.user_search = UserSearchBox(self)
        urwid.connect_signal(self.user_search, 'change',
                             self.update_user_list)
        self.view.user_search = self.user_search
        search_box = urwid.LineBox(
            self.user_search, tlcorner=u'─', tline=u'─', lline=u'',
            trcorner=u'─', blcorner=u'─', rline=u'',
            bline=u'─', brcorner=u'─'
            )
        self.search_lock = threading.Lock()
        super(RightColumnView, self).__init__(self.users_view(),
                                              header=search_box)

    @asyncn
    def update_user_list(self, search_box: Any, new_text: str) -> None:
        if not self.view.controller.editor_mode:
            return
        # wait for any previously started search to finish to avoid
        # displaying wrong user list.
        self.search_lock.acquire()
        users_display = self.users_btn_list.copy()
        for user in self.users_btn_list:
            if not match_user(user, new_text):
                users_display.remove(user)
        self.body = UsersView(
            urwid.SimpleFocusListWalker(users_display))
        self.set_body(self.body)
        self.view.controller.update_screen()
        self.search_lock.release()

    def users_view(self) -> Any:
        self.users_btn_list = list()  # type: List[Any]
        for user in self.view.users:
            unread_count = self.view.model.unread_counts.get(user['user_id'],
                                                             0)
            self.users_btn_list.append(
                UserButton(
                    user,
                    controller=self.view.controller,
                    view=self.view,
                    color=user['status'],
                    count=unread_count
                )
            )
        self.user_w = UsersView(
            urwid.SimpleFocusListWalker(self.users_btn_list))
        self.view.user_w = self.user_w
        return self.user_w

    def keypress(self, size: Tuple[int, int], key: str) -> str:
        if is_command_key('SEARCH_PEOPLE', key):
            self.set_focus('header')
            return key
        elif is_command_key('GO_BACK', key):
            self.user_search.set_edit_text("Search People")
            self.body = UsersView(
                urwid.SimpleFocusListWalker(self.users_btn_list))
            self.set_body(self.body)
            self.set_focus('body')
            self.view.controller.update_screen()
            return key
        return super(RightColumnView, self).keypress(size, key)


class LeftColumnView(urwid.Pile):
    """
    Displays the buttons at the left column of the app.
    """

    def __init__(self, view: Any) -> None:
        self.model = view.model
        self.view = view
        self.controller = view.controller
        left_column_structure = [
            (4, self.menu_view()),
            self.streams_view(),
        ]
        super(LeftColumnView, self).__init__(left_column_structure)

    def menu_view(self) -> Any:
        count = self.model.unread_counts.get('all_msg', 0)
        self.view.home_button = HomeButton(self.controller, count=count)
        count = self.model.unread_counts.get('all_pms', 0)
        self.view.pm_button = PMButton(self.controller, count=count)
        menu_btn_list = [
            self.view.home_button,
            self.view.pm_button,
        ]
        w = urwid.ListBox(urwid.SimpleFocusListWalker(menu_btn_list))
        return w

    def streams_view(self) -> Any:
        streams_btn_list = list()
        for stream in self.view.streams:
            unread_count = self.model.unread_counts.get(stream[1], 0)
            streams_btn_list.append(
                StreamButton(
                    stream,
                    controller=self.controller,
                    view=self.view,
                    count=unread_count,
                )
            )
        self.view.stream_w = StreamsView(streams_btn_list, self.view)
        w = urwid.LineBox(
            self.view.stream_w, title="Streams",
            tlcorner=u'─', tline=u'─', lline=u'',
            trcorner=u'─', blcorner=u'', rline=u'',
            bline=u'', brcorner=u'─'
            )
        return w

    def keypress(self, size: Tuple[int, int], key: str) -> str:
        if is_command_key('SEARCH_STREAMS', key):
            self.focus_position = 1
            self.view.stream_w.keypress(size, key)
            return key
        return super(LeftColumnView, self).keypress(size, key)


class HelpView(urwid.ListBox):
    def __init__(self, controller: Any) -> None:
        self.controller = controller
        self.log = urwid.SimpleFocusListWalker([
            urwid.Text("Press q to quit.", align='center')
        ])
        for _, binding in KEY_BINDINGS.items():
            commands = ", ".join(binding['keys'])
            self.log.append(
                urwid.Columns([
                    urwid.LineBox(
                        urwid.Text(binding['help_text']),
                        tlcorner=' ', tline=' ', lline=' ', trcorner=' ',
                        blcorner=' ', rline=' ', bline='-', brcorner=' '
                    ),
                    urwid.LineBox(
                        urwid.Text(commands),
                        tlcorner=' ', tline=' ', lline=' ', trcorner=' ',
                        blcorner=' ', rline=' ', bline='-', brcorner=' '
                    )
                ])
            )
        super(HelpView, self).__init__(self.log)

    def keypress(self, size: Tuple[int, int], key: str) -> str:
        if is_command_key('QUIT_HELP', key):
            self.controller.exit_help()
        return super(HelpView, self).keypress(size, key)



import time
from collections import defaultdict
from functools import wraps
from threading import Thread
from typing import Any, Dict, List

import os


def asyncn(func: Any) -> Any:
    """
    Decorator for executing a function in a separate :class:`threading.Thread`.
    """
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        # If calling when pytest is running simply return the function
        # to avoid running in asyncn mode.
        if os.environ.get("PYTEST_CURRENT_TEST"):
            return func(*args, **kwargs)

        thread = Thread(target=func, args=args, kwargs=kwargs)
        thread.daemon = True
        return thread.start()
    return wrapper


def set_count(id_list: List[int], controller: Any, new_count: int) -> None:
    messages = controller.model.index['messages']
    for id in id_list:
        msg = messages[id]
        msg_type = msg['type']
        if msg_type == 'stream':
            unread_id = messages[id]['stream_id']
            stream_topic = (unread_id, msg['subject'])
            unread_counts = controller.model.unread_counts
            if stream_topic in unread_counts['unread_topics'].keys():
                unread_counts['unread_topics'][stream_topic] += new_count
                if unread_counts['unread_topics'][stream_topic] == 0:
                    unread_counts['unread_topics'].pop(stream_topic)
            elif new_count == 1:
                unread_counts['unread_topics'][stream_topic] = new_count
        else:
            unread_id = messages[id]['sender_id']
            unread_counts = controller.model.unread_counts
            if unread_id in unread_counts['unread_pms'].keys():
                unread_counts['unread_pms'][unread_id] += new_count
                if unread_counts['unread_pms'][unread_id] == 0:
                    unread_counts['unread_pms'].pop(unread_id)
            elif new_count == 1:
                unread_counts['unread_pms'][unread_id] = new_count

    # if view is not yet loaded. Usually the case when first message is read.
    while not hasattr(controller, 'view'):
        time.sleep(0.1)

    streams = controller.view.stream_w.log
    users = controller.view.user_w.log
    all_msg = controller.view.home_button
    all_pm = controller.view.pm_button
    for id in id_list:
        msg_type = messages[id]['type']
        if msg_type == 'stream':
            stream_id = messages[id]['stream_id']
            for stream in streams:
                if stream.stream_id == stream_id:
                    stream.update_count(stream.count + new_count)
                    break

        else:
            user_id = messages[id]['sender_id']
            for user in users:
                if user.user_id == user_id:
                    user.update_count(user.count + new_count)
                    break
            all_pm.update_count(all_pm.count + new_count)
        all_msg.update_count(all_msg.count + new_count)

    while not hasattr(controller, 'loop'):
        time.sleep(0.1)
    controller.update_screen()


@asyncn
def update_flag(id_list: List[int], controller: Any) -> None:
    if not id_list:
        return
    request = {
        'messages': id_list,
        'flag': 'read',
        'op': 'add',
    }
    client = controller.client
    client.do_api_query(request, '/json/messages/flags', method="POST")
    set_count(id_list, controller, -1)


def index_messages(messages: List[Any], model: Any, index: Any=None)\
        -> Dict[str, Any]:
    """
    STRUCTURE OF INDEX
    {
        'pointer': {
            '[]': 30  # str(ZulipModel.narrow)
            '[["stream", "verona"]]': 32,
            ...
        }
        'stream': {
            123: {    # stream_id
                'topic name': {
                    51234,  # message id
                    56454,
                    ...
                }
        },
        'private': {
            (3, 7): {  # user_ids frozenset
                51234,
                56454,
                ...
            },
            (1, 2, 3, 4): {  # multiple recipients
                12345,
                32553,
            }
        },
        'all_messages': {
            14231,
            23423,
            ...
        },
        'all_private': {
            22334,
            23423,
            ...
        },
        'all_stream': {
            123: {
                53434,
                36435,
                ...
            }
            234: {
                23423,
                23423,
                ...
            }
        }
        'search': {
            13242,
            23423,
            23423,
            ...
        },
        'messages': {
            # all the messages mapped to their id
            # for easy retrieval of message from id
            45645: {  # PRIVATE
                'id': 4290,
                'timestamp': 1521817473,
                'content': 'Hi @**Cordelia Lear**',
                'sender_full_name': 'Iago',
                'flags': [],
                'sender_short_name': 'iago',
                'sender_email': 'iago@zulip.com',
                'subject': '',
                'subject_links': [],
                'sender_id': 73,
                'type': 'private',
                'recipient_id': 124,
                'reactions': [],
                'display_recipient': [
                    {
                        'email': 'ZOE@zulip.com',
                        'id': 70,
                        'full_name': 'Zoe',
                    }, {
                        'email': 'cordelia@zulip.com',
                        'id': 71,
                        'full_name': 'Cordelia Lear',
                    }, {
                        'email': 'hamlet@zulip.com',
                        'id': 72,
                        'full_name': 'King Hamlet',
                    }, {
                        'email': 'iago@zulip.com',
                        'id': 73,
                        'full_name': 'Iago',
                    }
                ]
            },
            45645: {  # STREAM
                'timestamp': 1521863062,
                'sender_id': 72,
                'sender_full_name': 'King Hamlet',
                'recipient_id': 119,
                'content': 'https://github.com/zulip/zulip-terminal',
                'type': 'stream',
                'sender_email': 'hamlet@zulip.com',
                'id': 4298,
                'display_recipient': 'Verona',
                'flags': [],
                'reactions': [],
                'subject': 'Verona2',
                'stream_id': 32,
            },
        },
    }
    """
    if index is None:
        index = {
            'pointer': defaultdict(set),
            'stream': defaultdict(dict),
            'private': defaultdict(set),
            'all_messages': set(),
            'all_private': set(),
            'all_stream': defaultdict(set),
            'messages': defaultdict(dict),
            'search': set(),
        }
    narrow = model.narrow
    for msg in messages:

        index['messages'][msg['id']] = msg
        if not narrow:
            index['all_messages'].add(msg['id'])

        elif narrow[0][0] == 'search':
            index['search'].add(msg['id'])
            continue

        if len(narrow) == 1:

            if msg['type'] == 'private':
                index['all_private'].add(msg['id'])
                recipients = frozenset(set(
                    recipient['id'] for recipient in msg['display_recipient']
                ))

                if narrow[0][0] == 'pm_with' and recipients == frozenset({
                        model.user_id, model.user_dict[narrow[0][1]]['user_id']
                }):
                    index['private'][recipients].add(msg['id'])

            if msg['type'] == 'stream' and msg['stream_id'] == model.stream_id:
                index['all_stream'][msg['stream_id']].add(msg['id'])

        if msg['type'] == 'stream' and len(narrow) == 2 and\
                narrow[1][1] == msg['subject']:

            if not index['stream'][msg['stream_id']].get(msg['subject']):
                index['stream'][msg['stream_id']][msg['subject']] = set()
            index['stream'][msg['stream_id']][msg['subject']].add(msg['id'])

    return index


def classify_unread_counts(model: Any) -> Dict[str, Any]:
    # TODO: supprot group pms
    unread_msg_counts = model.initial_data['unread_msgs']
    unread_counts = dict()  # type: Dict[Any, Any]
    unread_counts['all_msg'] = 0
    unread_counts['all_pms'] = 0
    unread_counts['unread_topics'] = dict()
    unread_counts['unread_pms'] = dict()

    for pm in unread_msg_counts['pms']:
        count = len(pm['unread_message_ids'])
        unread_counts[pm['sender_id']] = count
        unread_counts['unread_pms'][pm['sender_id']] = count
        unread_counts['all_msg'] += count
        unread_counts['all_pms'] += count

    for stream in unread_msg_counts['streams']:
        count = len(stream['unread_message_ids'])
        stream_id = stream['stream_id']
        if stream_id in model.muted_streams:
            continue
        if [model.stream_dict[stream_id]['name'],
                stream['topic']] in model.muted_topics:
            continue
        stream_topic = (stream_id, stream['topic'])
        unread_counts['unread_topics'][stream_topic] = count
        if not unread_counts.get(stream_id):
            unread_counts[stream_id] = count
        else:
            unread_counts[stream_id] += count
        unread_counts['all_msg'] += count

    return unread_counts


def match_user(user: Any, text: str) -> bool:
    """
    Matches if the user full name, last name or email matches
    with `text` or not.
    """
    full_name = user.caption.lower()
    keywords = full_name.split()
    keywords.append(user.email.lower())
    # adding full_name helps in further narrowing down the right user.
    keywords.append(full_name)
    for keyword in keywords:
        if keyword.startswith(text.lower()):
            return True
    return False



import pytest

from zulipterminal.ui_tools.views import (
    MessageView,
    MiddleColumnView,
    StreamsView,
    UsersView,
    RightColumnView,
    LeftColumnView,
    HelpView,
)

VIEWS = "zulipterminal.ui_tools.views"


class TestMessageView:

    @pytest.fixture(autouse=True)
    def mock_external_classes(self, mocker):
        self.model = mocker.MagicMock()
        self.urwid = mocker.patch(VIEWS + ".urwid")

    @pytest.fixture
    def msg_view(self, mocker, msg_box):
        mocker.patch(VIEWS + ".MessageView.main_view", return_value=[msg_box])
        mocker.patch(VIEWS + ".MessageView.read_message")
        self.urwid.SimpleFocusListWalker.return_value = mocker.Mock()
        mocker.patch(VIEWS + ".MessageView.set_focus")
        msg_view = MessageView(self.model)
        return msg_view

    def test_init(self, mocker, msg_view, msg_box):
        assert msg_view.model == self.model
        self.urwid.SimpleFocusListWalker.assert_called_once_with([msg_box])
        msg_view.set_focus.assert_called_once_with(0)
        assert msg_view.old_loading is False
        assert msg_view.new_loading is False

    @pytest.mark.parametrize("narrow_focus_pos, focus_msg", [
        (set(), 1), (0, 0)
    ])
    def test_main_view(self, mocker, narrow_focus_pos, focus_msg):
        mocker.patch(VIEWS + ".MessageView.read_message")
        self.urwid.SimpleFocusListWalker.return_value = mocker.Mock()
        mocker.patch(VIEWS + ".MessageView.set_focus")
        msg_list = ["MSG1", "MSG2"]
        mocker.patch(VIEWS + ".create_msg_box_list",
                     return_value=msg_list)
        self.model.get_focus_in_current_narrow.return_value = narrow_focus_pos
        msg_view = MessageView(self.model)
        assert msg_view.focus_msg == focus_msg

    def test_load_old_messages(self, mocker, msg_view):
        mocker.patch.object(msg_view.model,
                            "get_message_ids_in_current_narrow",
                            return_value={0})
        self.model.get_messages.return_value = {}
        create_msg_box_list = mocker.patch(VIEWS + ".create_msg_box_list",
                                           return_value=["M1", "M2"])
        msg_view.load_old_messages(0)
        assert msg_view.old_loading is False
        assert msg_view.model.anchor == 0
        assert msg_view.model.num_after == 0
        assert msg_view.model.num_before == 30
        assert msg_view.index == {}
        create_msg_box_list.assert_called_once_with(msg_view.model, set())
        self.model.controller.update_screen.assert_called_once_with()

    def test_load_new_messages(self, mocker, msg_view):
        mocker.patch.object(msg_view.model,
                            "get_message_ids_in_current_narrow",
                            return_value={0})
        self.model.get_messages.return_value = {}
        create_msg_box_list = mocker.patch(VIEWS + ".create_msg_box_list",
                                           return_value=["M1", "M2"])
        msg_view.load_new_messages(0)
        assert msg_view.new_loading is False
        assert msg_view.model.anchor == 0
        assert msg_view.model.num_after == 30
        assert msg_view.model.num_before == 0
        assert msg_view.index == {}
        msg_view.log.extend.assert_called_once_with(['M1', 'M2'])
        create_msg_box_list.assert_called_once_with(msg_view.model, set())
        self.model.controller.update_screen.assert_called_once_with()

    @pytest.mark.parametrize("event, button, keypress", [
        ("mouse press", 4, "up"),
        ("mouse press", 5, "down"),
    ])
    def test_mouse_event(self, mocker, msg_view, event, button, keypress):
        mocker.patch.object(msg_view, "keypress")
        msg_view.mouse_event((20,), event, button, 0, 0, mocker.Mock())
        msg_view.keypress.assert_called_once_with((20,), keypress)

    # down key tests
    def test_keypress_down_key(self, mocker, msg_view):
        size = (20,)
        key = "down"
        msg_view.new_loading = False
        mocker.patch(VIEWS + ".MessageView.focus_position", return_value=0)
        msg_view.log.next_position.return_value = 1
        msg_view.keypress(size, key)
        msg_view.log.next_position.assert_called_once_with(
            msg_view.focus_position)
        msg_view.set_focus.assert_called_with(1, 'above')

    def test_keypress_down_key_exception(self, mocker, msg_view):
        size = (20,)
        key = "down"
        msg_view.new_loading = False
        mocker.patch(VIEWS + ".MessageView.focus_position", return_value=0)

        # Raise exception
        msg_view.log.next_position = Exception()
        mocker.patch(VIEWS + ".MessageView.focus", return_value=True)
        mocker.patch.object(msg_view, "load_new_messages")
        msg_view.keypress(size, key)
        msg_view.load_new_messages.assert_called_once_with(
            msg_view.focus.original_widget.message['id'],
        )

        # No message in focus
        msg_view.focus = False
        return_value = msg_view.keypress(size, key)
        assert return_value == key

    # up key tests
    def test_keypress_up(self, mocker, msg_view):
        key = "up"
        size = (20,)
        mocker.patch(VIEWS + ".MessageView.focus_position", return_value=0)
        msg_view.old_loading = False
        msg_view.log.prev_position.return_value = 1
        msg_view.keypress(size, key)
        msg_view.log.prev_position.assert_called_once_with(
            msg_view.focus_position)
        msg_view.set_focus.assert_called_with(1, 'below')

    def test_keypress_up_raise_exception(self, mocker, msg_view):
        key = "up"
        size = (20,)
        mocker.patch(VIEWS + ".MessageView.focus_position", return_value=0)
        msg_view.old_loading = False

        # Raise exception
        msg_view.log.prev_position = Exception()
        mocker.patch(VIEWS + ".MessageView.focus", return_value=True)
        mocker.patch.object(msg_view, "load_old_messages")

        msg_view.keypress(size, key)
        msg_view.load_old_messages.assert_called_once_with(
            msg_view.focus.original_widget.message['id'],
        )

        # No message in focus
        msg_view.focus = False
        return_value = msg_view.keypress(size, key)
        msg_view.load_old_messages.assert_called_with()
        assert return_value == key

    def test_read_message(self, mocker, msg_box):
        mocker.patch(VIEWS + ".MessageView.main_view", return_value=[msg_box])
        self.urwid.SimpleFocusListWalker.return_value = mocker.Mock()
        mocker.patch(VIEWS + ".MessageView.set_focus")
        mocker.patch(VIEWS + ".MessageView.update_current_footer")
        msg_view = MessageView(self.model)
        msg_w = mocker.MagicMock()
        msg_w.attr_map = {None: 'unread'}
        msg_w.original_widget.message = {'id': 1}
        msg_w.set_attr_map.return_value = None
        msg_view.body.get_focus.return_value = (msg_w, 0)
        msg_view.body.get_prev.return_value = (None, 1)
        update_flag = mocker.patch(VIEWS + ".update_flag")
        msg_view.model.narrow = []
        msg_view.model.index = {
            'messages': {
                1: {
                    'flags': [],
                }
            },
            'pointer': {
                '[]': 0
            }
        }
        mocker.patch(VIEWS + ".MessageView.focus_position")
        msg_view.focus_position = 1
        msg_view.read_message()
        assert msg_view.update_current_footer.called
        assert msg_view.model.index['messages'][1]['flags'] == ['read']
        update_flag.assert_called_once_with([1], self.model.controller)

    def test_read_message_no_msgw(self, mocker, msg_view):
        # MSG_W is NONE CASE
        msg_view.body.get_focus.return_value = (None, 0)
        update_flag = mocker.patch(VIEWS + ".update_flag")
        msg_view.read_message()
        update_flag.assert_not_called()


class TestStreamsView:

    @pytest.fixture
    def stream_view(self, mocker):
        self.log = mocker.patch(VIEWS + ".urwid.SimpleFocusListWalker",
                                return_value=[])
        mocker.patch(VIEWS + ".urwid.connect_signal")
        mocker.patch(VIEWS + ".threading.Lock")
        self.view = mocker.Mock()
        self.search_box = mocker.patch(VIEWS + ".StreamSearchBox")
        stream_btn = mocker.Mock()
        stream_btn.caption = "FOO"
        self.streams_btn_list = [stream_btn]
        return StreamsView(self.streams_btn_list, view=self.view)

    def test_init(self, mocker, stream_view):
        assert stream_view.view == self.view
        assert stream_view.log == []
        assert stream_view.streams_btn_list == self.streams_btn_list
        assert stream_view.search_box
        self.search_box.assert_called_once_with(stream_view)

    def test_update_streams(self, mocker, stream_view):
        self.view.controller.editor_mode = True
        new_text = "F"
        search_box = "SEARCH_BOX"
        stream_view.update_streams(search_box, new_text)
        assert not stream_view.log

    def test_mouse_event(self, mocker, stream_view):
        mocker.patch.object(stream_view, 'keypress')
        size = (200, 20)
        col = 1
        row = 1
        focus = "WIDGET"
        # Left click
        stream_view.mouse_event(size, "mouse press", 4, col, row, focus)
        stream_view.keypress.assert_called_once_with(size, "up")

        # Right click
        stream_view.mouse_event(size, "mouse press", 5, col, row, focus)
        stream_view.keypress.assert_called_with(size, "down")

    def test_keypress_q(self, mocker, stream_view):
        key = "q"
        size = (20,)
        mocker.patch.object(stream_view, 'set_focus')
        stream_view.keypress(size, key)
        stream_view.set_focus.assert_called_once_with("header")

    def test_keypress_esc(self, mocker, stream_view):
        key = "esc"
        size = (20,)
        mocker.patch.object(stream_view, 'set_focus')
        stream_view.keypress(size, key)
        stream_view.set_focus.assert_called_once_with("body")
        stream_view.search_box.set_edit_text.assert_called_once_with(
            "Search streams")
        assert stream_view.log == self.streams_btn_list


class TestUsersView:

    @pytest.fixture
    def user_view(self, mocker):
        mocker.patch(VIEWS + ".urwid.SimpleFocusListWalker", return_value=[])
        return UsersView("USER_BTN_LIST")

    def test_mouse_event(self, mocker, user_view):
        mocker.patch.object(user_view, 'keypress')
        size = (200, 20)
        col = 1
        row = 1
        focus = "WIDGET"
        # Left click
        user_view.mouse_event(size, "mouse press", 4, col, row, focus)
        user_view.keypress.assert_called_with(size, "up")

        # Right click
        user_view.mouse_event(size, "mouse press", 5, col, row, focus)
        user_view.keypress.assert_called_with(size, "down")

        # Other actions - No action
        return_value = user_view.mouse_event(
            size, "mouse release", 4, col, row, focus)
        assert return_value is False

        # Other clicks
        return_value = user_view.mouse_event(
            size, "mouse press", 1, col, row, focus)
        assert return_value is False


class TestMiddleColumnView:

    @pytest.fixture(autouse=True)
    def mock_external_classes(self, mocker):
        mocker.patch(VIEWS + ".MessageView", return_value="MSG_LIST")
        self.model = mocker.Mock()
        self.write_box = mocker.Mock()
        self.search_box = mocker.Mock()
        self.super = mocker.patch(VIEWS + '.urwid.Frame.__init__')
        self.super_keypress = mocker.patch(VIEWS + '.urwid.Frame.keypress')
        self.model.controller == mocker.Mock()

    @pytest.fixture
    def mid_col_view(self):
        return MiddleColumnView(self.model, self.write_box, self.search_box)

    def test_init(self, mid_col_view):
        assert mid_col_view.model == self.model
        assert mid_col_view.controller == self.model.controller
        assert mid_col_view.last_unread_topic is None
        assert mid_col_view.last_unread_pm is None
        assert mid_col_view.search_box == self.search_box
        assert self.model.msg_list == "MSG_LIST"
        self.super.assert_called_once_with("MSG_LIST", header=self.search_box,
                                           footer=self.write_box)

    def test_get_next_unread_topic(self, mid_col_view):
        mid_col_view.model.unread_counts = {
            'unread_topics': {1: 1, 2: 1}
        }
        return_value = mid_col_view.get_next_unread_topic()
        assert return_value == 1
        assert mid_col_view.last_unread_topic == 1

    def test_get_next_unread_topic_again(self, mid_col_view):
        mid_col_view.model.unread_counts = {
            'unread_topics': {1: 1, 2: 1}
        }
        mid_col_view.last_unread_topic = 1
        return_value = mid_col_view.get_next_unread_topic()
        assert return_value == 2
        assert mid_col_view.last_unread_topic == 2

    def test_get_next_unread_topic_no_unread(self, mid_col_view):
        mid_col_view.model.unread_counts = {
            'unread_topics': {}
        }
        return_value = mid_col_view.get_next_unread_topic()
        assert return_value is None
        assert mid_col_view.last_unread_topic is None

    def test_get_next_unread_pm(self, mid_col_view):
        mid_col_view.model.unread_counts = {
            'unread_pms': {1: 1, 2: 1}
        }
        return_value = mid_col_view.get_next_unread_pm()
        assert return_value == 1
        assert mid_col_view.last_unread_pm == 1

    def test_get_next_unread_pm_again(self, mid_col_view):
        mid_col_view.model.unread_counts = {
            'unread_pms': {1: 1, 2: 1}
        }
        mid_col_view.last_unread_pm = 1
        return_value = mid_col_view.get_next_unread_pm()
        assert return_value == 2
        assert mid_col_view.last_unread_pm == 2

    def test_get_next_unread_pm_no_unread(self, mid_col_view):
        mid_col_view.model.unread_counts = {
            'unread_pms': {}
        }
        return_value = mid_col_view.get_next_unread_pm()
        assert return_value is None
        assert mid_col_view.last_unread_pm is None

    def test_keypress_esc(self, mid_col_view, mocker):
        key = "esc"
        size = (20,)
        mocker.patch(VIEWS + '.MiddleColumnView.header')
        mocker.patch(VIEWS + '.MiddleColumnView.footer')
        mocker.patch(VIEWS + '.MiddleColumnView.set_focus')

        mid_col_view.keypress(size, key)

        mid_col_view.header.keypress.assert_called_once_with(size, key)
        mid_col_view.footer.keypress.assert_called_once_with(size, key)
        mid_col_view.set_focus.assert_called_once_with('body')
        self.super_keypress.assert_called_once_with(size, key)

    def test_keypress_focus_header(self, mid_col_view, mocker):
        key = "/"
        size = (20,)
        mid_col_view.focus_part = 'header'
        mid_col_view.keypress(size, key)
        self.super_keypress.assert_called_once_with(size, key)

    def test_keypress_search(self, mid_col_view, mocker):
        key = "/"
        size = (20,)
        mocker.patch(VIEWS + '.MiddleColumnView.focus_position')
        mocker.patch(VIEWS + '.MiddleColumnView.set_focus')

        mid_col_view.keypress(size, key)

        assert mid_col_view.controller.editor_mode is True
        assert mid_col_view.controller.editor == mid_col_view.search_box
        mid_col_view.set_focus.assert_called_once_with('header')

    def test_keypress_r(self, mid_col_view, mocker):
        key = "r"
        size = (20,)
        mocker.patch(VIEWS + '.MiddleColumnView.body')
        mocker.patch(VIEWS + '.MiddleColumnView.footer')
        mocker.patch(VIEWS + '.MiddleColumnView.focus_position')
        mocker.patch(VIEWS + '.MiddleColumnView.set_focus')

        mid_col_view.keypress(size, key)

        mid_col_view.body.keypress.assert_called_once_with(size, 'enter')
        mid_col_view.set_focus.assert_called_once_with('footer')
        assert mid_col_view.footer.focus_position == 1

    def test_keypress_c(self, mid_col_view, mocker):
        key = "c"
        size = (20,)
        mocker.patch(VIEWS + '.MiddleColumnView.body')
        mocker.patch(VIEWS + '.MiddleColumnView.footer')
        mocker.patch(VIEWS + '.MiddleColumnView.focus_position')
        mocker.patch(VIEWS + '.MiddleColumnView.set_focus')

        mid_col_view.keypress(size, key)

        mid_col_view.body.keypress.assert_called_once_with(size, 'c')
        mid_col_view.set_focus.assert_called_once_with('footer')
        assert mid_col_view.footer.focus_position == 0

    def test_keypress_R(self, mid_col_view, mocker):
        key = "R"
        size = (20,)
        mocker.patch(VIEWS + '.MiddleColumnView.body')
        mocker.patch(VIEWS + '.MiddleColumnView.footer')
        mocker.patch(VIEWS + '.MiddleColumnView.focus_position')
        mocker.patch(VIEWS + '.MiddleColumnView.set_focus')

        mid_col_view.keypress(size, key)

        mid_col_view.body.keypress.assert_called_once_with(size, 'R')
        mid_col_view.set_focus.assert_called_once_with('footer')
        assert mid_col_view.footer.focus_position == 1

    def test_keypress_n_stream(self, mid_col_view, mocker):
        key = "n"
        size = (20,)
        mocker.patch(VIEWS + '.MiddleColumnView.focus_position')
        topic_btn = mocker.patch(VIEWS + '.TopicButton')
        mocker.patch(VIEWS + '.MiddleColumnView.get_next_unread_topic',
                     return_value=('stream', 'topic'))

        mid_col_view.keypress(size, key)

        mid_col_view.get_next_unread_topic.assert_called_once_with()
        mid_col_view.controller.narrow_to_topic.assert_called_once_with(
            topic_btn('stream', 'topic', mid_col_view.model)
        )

    def test_keypress_n_no_stream(self, mid_col_view, mocker):
        key = "n"
        size = (20,)
        mocker.patch(VIEWS + '.MiddleColumnView.focus_position')
        topic_btn = mocker.patch(VIEWS + '.TopicButton')
        mocker.patch(VIEWS + '.MiddleColumnView.get_next_unread_topic',
                     return_value=None)

        return_value = mid_col_view.keypress(size, key)
        assert return_value == 'n'

    def test_keypress_p_stream(self, mid_col_view, mocker):
        key = "p"
        size = (20,)
        mocker.patch(VIEWS + '.MiddleColumnView.focus_position')
        pm_btn = mocker.patch(VIEWS + '.UnreadPMButton')
        mocker.patch(VIEWS + '.MiddleColumnView.get_next_unread_pm',
                     return_value=1)
        mid_col_view.model.user_id_email_dict = {1: "EMAIL"}

        mid_col_view.keypress(size, key)

        mid_col_view.controller.narrow_to_user.assert_called_once_with(
            pm_btn(1, 'EMAIL')
        )

    def test_keypress_p_no_pm(self, mid_col_view, mocker):
        key = "p"
        size = (20,)
        mocker.patch(VIEWS + '.MiddleColumnView.focus_position')
        pm_btn = mocker.patch(VIEWS + '.UnreadPMButton')
        mocker.patch(VIEWS + '.MiddleColumnView.get_next_unread_pm',
                     return_value=None)

        return_value = mid_col_view.keypress(size, key)
        assert return_value == 'p'

    def test_keypress_x(self, mid_col_view, mocker):
        key = "x"
        size = (20,)
        mocker.patch(VIEWS + '.MiddleColumnView.focus_position')
        pm_btn = mocker.patch(VIEWS + '.UnreadPMButton')
        mocker.patch(VIEWS + '.MiddleColumnView.get_next_unread_pm',
                     return_value=None)
        mid_col_view.footer = mocker.Mock()
        return_value = mid_col_view.keypress(size, key)
        mid_col_view.footer.private_box_view.assert_called_once_with()
        assert mid_col_view.footer.focus_position == 0
        assert return_value == 'x'


class TestRightColumnView:

    @pytest.fixture(autouse=True)
    def mock_external_classes(self, mocker):
        self.view = mocker.Mock()
        self.user_search = mocker.patch(VIEWS + ".UserSearchBox")
        self.connect_signal = mocker.patch(VIEWS + ".urwid.connect_signal")
        self.line_box = mocker.patch(VIEWS + ".urwid.LineBox")
        self.thread = mocker.patch(VIEWS + ".threading")
        self.super = mocker.patch(VIEWS + ".urwid.Frame.__init__")

    @pytest.fixture
    def right_col_view(self, mocker):
        mocker.patch(VIEWS + ".RightColumnView.users_view")
        return RightColumnView(self.view)

    def test_init(self, right_col_view):
        assert right_col_view.view == self.view
        assert right_col_view.user_search == self.user_search(right_col_view)
        assert right_col_view.view.user_search == right_col_view.user_search
        self.thread.Lock.assert_called_with()
        assert right_col_view.search_lock == self.thread.Lock()
        self.super.assert_called_once_with(right_col_view.users_view(),
                                           header=self.line_box(
                                               right_col_view.user_search
        ))

    def test_update_user_list_editor_mode(self, right_col_view):
        right_col_view.view.controller.editor_mode = False
        right_col_view.update_user_list("SEARCH_BOX", "NEW_TEXT")
        right_col_view.search_lock.acquire.assert_not_called()

    def test_update_user_list_user_match(self, right_col_view, mocker):
        right_col_view.view.controller.editor_mode = True
        right_col_view.users_btn_list = ["USER1", "USER2"]
        mocker.patch(VIEWS + ".match_user", return_value=True)
        mocker.patch(VIEWS + ".UsersView")
        list_w = mocker.patch(VIEWS + ".urwid.SimpleFocusListWalker")
        set_body = mocker.patch(VIEWS + ".urwid.Frame.set_body")

        right_col_view.update_user_list("SEARCH_BOX", "F")

        right_col_view.search_lock.acquire.assert_called_once_with()
        list_w.assert_called_once_with(["USER1", "USER2"])
        set_body.assert_called_once_with(right_col_view.body)

    def test_update_user_list_no_user_match(self, right_col_view, mocker):
        right_col_view.view.controller.editor_mode = True
        right_col_view.users_btn_list = ["USER1", "USER2"]
        mocker.patch(VIEWS + ".match_user", return_value=False)
        mocker.patch(VIEWS + ".UsersView")
        list_w = mocker.patch(VIEWS + ".urwid.SimpleFocusListWalker")
        set_body = mocker.patch(VIEWS + ".urwid.Frame.set_body")

        right_col_view.update_user_list("SEARCH_BOX", "F")

        right_col_view.search_lock.acquire.assert_called_once_with()
        list_w.assert_called_once_with([])
        set_body.assert_called_once_with(right_col_view.body)

    def test_users_view(self, mocker):
        self.view.users = [{
            'user_id': 1,
            'status': 'active'
        }]
        self.view.model.unread_counts.get.return_value = 1
        user_btn = mocker.patch(VIEWS + ".UserButton")
        mocker.patch(VIEWS + ".UsersView")
        list_w = mocker.patch(VIEWS + ".urwid.SimpleFocusListWalker")

        right_col_view = RightColumnView(self.view)

        right_col_view.view.model.unread_counts.get.assert_called_once_with(1,
                                                                            0)
        user_btn.assert_called_once_with(
            self.view.users[0],
            controller=self.view.controller,
            view=self.view,
            color=self.view.users[0]['status'],
            count=1
        )
        list_w.assert_called_once_with(right_col_view.users_btn_list)

    def test_keypress_w(self, right_col_view, mocker):
        key = 'w'
        size = (20,)
        mocker.patch(VIEWS + ".RightColumnView.set_focus")
        right_col_view.keypress(size, key)
        right_col_view.set_focus.assert_called_once_with('header')

    def test_keypress_esc(self, right_col_view, mocker):
        key = 'esc'
        size = (20,)
        mocker.patch(VIEWS + ".UsersView")
        list_w = mocker.patch(VIEWS + ".urwid.SimpleFocusListWalker")
        mocker.patch(VIEWS + ".RightColumnView.set_focus")
        mocker.patch(VIEWS + ".RightColumnView.set_body")
        right_col_view.users_btn_list = []

        right_col_view.keypress(size, key)

        right_col_view.user_search.set_edit_text.assert_called_once_with(
            "Search People"
        )
        right_col_view.set_body.assert_called_once_with(right_col_view.body)
        right_col_view.set_focus.assert_called_once_with('body')
        list_w.assert_called_once_with([])


class TestLeftColumnView:
    @pytest.fixture(autouse=True)
    def mock_external_classes(self, mocker):
        self.view = mocker.Mock()
        self.view.model = mocker.Mock()
        self.view.model.unread_counts = {
            'all_msg': 2,
            'all_pms': 0,
            86: 1,
            14: 1,
        }
        self.view.controller = mocker.Mock()
        self.super_mock = mocker.patch(VIEWS + ".urwid.Pile.__init__")

    @pytest.fixture
    def left_col_view(self, mocker):
        self.menu_view = mocker.patch(VIEWS + ".LeftColumnView.menu_view")
        self.streams_view = mocker.patch(
            VIEWS + ".LeftColumnView.streams_view")
        return LeftColumnView(self.view)

    def test_menu_view(self, mocker):
        self.streams_view = mocker.patch(
            VIEWS + ".LeftColumnView.streams_view")
        home_button = mocker.patch(VIEWS + ".HomeButton")
        pm_button = mocker.patch(VIEWS + ".PMButton")
        mocker.patch(VIEWS + ".urwid.ListBox")
        mocker.patch(VIEWS + ".urwid.SimpleFocusListWalker")
        left_col_view = LeftColumnView(self.view)
        home_button.assert_called_once_with(left_col_view.controller,
                                            count=2)
        pm_button.assert_called_once_with(left_col_view.controller,
                                          count=0)

    def test_streams_view(self, mocker, streams):
        self.view.streams = streams
        stream_button = mocker.patch(VIEWS + '.StreamButton')
        stream_view = mocker.patch(VIEWS + '.StreamsView')
        line_box = mocker.patch(VIEWS + '.urwid.LineBox')
        left_col_view = LeftColumnView(self.view)
        stream_button.assert_called_with(
            streams[1],
            controller=self.view.controller,
            view=self.view,
            count=1)


class TestHelpMenu:
    @pytest.fixture(autouse=True)
    def mock_external_classes(self, mocker, monkeypatch):
        self.controller = mocker.Mock()
        mocker.patch(VIEWS + ".urwid.SimpleFocusListWalker", return_value=[])
        self.help_view = HelpView(self.controller)

    def test_keypress_any_key(self):
        key = "a"
        size = (200, 20)
        self.help_view.keypress(size, key)
        assert not self.controller.exit_help.called

    def test_keypress_q(self):
        key = "q"
        size = (200, 20)
        self.help_view.keypress(size, key)
        assert self.controller.exit_help.called
